# -*- coding: utf-8 -*-
#
# Created : 2015-04-03
#
# @author: Eric Lapouyade
#
"""This module defines base classes for python-textops3"""

import os
import sys
import re
import types
import textops
from addicted import NoAttrDict, NoAttr
import string
import logging
import pprint
import chardet
from collections import abc
pp = pprint.PrettyPrinter(indent=4)


logger = textops.logger

def activate_debug():
    """Activate debug logging on console

    This function is useful when playing with python-textops3 through a python console.
    It is not recommended to use this function in a real application : use standard logging
    functions instead.
    """
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    logger.addHandler(ch)
    logger.setLevel(logging.DEBUG)

class TextOpException(Exception):
    pass

class TextOp(object):
    """Base class for text operations

    All operations must be derived from this class. Subclasses must redefine an ``op()`` method
    that will be called when the operations will be triggered by an input text.
    """
    def __init__(self,*args,**kwargs):
        self.ops = [[self.__class__.__name__, args, kwargs]]
        self.op = None
        self.debug = kwargs.get('debug',False)

    def __getattr__(self,attr):
        if not attr.startswith('_'):
            self.ops.append([attr, (), {}])
            self.op = attr
        else:
            raise AttributeError()
        return self

    def __or__(self,other):
        if not isinstance(other,TextOp):
            raise TextOpException('Please use "|" only between two TextOp or AFTER a string or a list')
        self.ops += other.ops
        return self

    def __ror__(self,text):
        return self._process(text, piped=True)

    def __rrshift__(self,text):
        result = self._process(text, piped=True)
        if isinstance(result, (types.GeneratorType,enumerate)):
            return ListExt(result)
        return result

    def __rshift__(self,other):
        if not isinstance(other,TextOp):
            raise TextOpException('Please use ">>" only between two TextOp or AFTER a string or a list')
        self.ops += other.ops
        return self

    def __str__(self):
        return self.se

    def __iter__(self):
        return iter(self.g)

    def __int__(self):
        return iter(self.i)

    def __float__(self):
        return iter(self.f)

    def __call__(self,*args,**kwargs):
        if self.op:
            self.ops[-1][1] = args
            self.ops[-1][2] = kwargs
            self.op = None
            return self
        else:
            return self._process(args and args[0] or None)

    def _process(self,text=None, piped=False):
        input_text = text
        if self.debug:
            if isinstance(text, types.GeneratorType):
                text = list(text)
            logger.debug('=== TextOps : %r' % self)
            logger.debug(DebugText(text))
        for i,(op,args,kwargs) in enumerate(self.ops):
            if not piped and not i and not input_text and args:
                text = args[0]
                args = args[1:]
            opcls = getattr(textops.ops,op,None)
            if isinstance(opcls,type) and issubclass(opcls, TextOp):
                try:
                    text = opcls.op(text, *args, **kwargs)
                    if self.debug:
                        if isinstance(text, types.GeneratorType):
                            text = list(text)
                        logger.debug('--- Op : %s(%s,%s)',op,args,kwargs)
                        logger.debug(DebugText(text))
                    if text is None:
                         return text
                except TypeError:
                    logger.error('*** bad parameters for %s()' % opcls.__name__)
                    raise
            elif hasattr(text,op):
                text = getattr(text,op)(*args, **kwargs)
            else:
                extext = extend_type(text)
                if hasattr(extext,op):
                    text = getattr(extext,op)(*args, **kwargs)
                elif isinstance(text, (types.GeneratorType,enumerate,list)):
                    text = self._apply_op_gen(text,op,*args, **kwargs)
                else:
                    raise TextOpException('Unknown OP "%s"' % op)

        return extend_type(text)

    def _apply_op_gen(self, text, op, *args, **kwargs):
        for line in text:
            if hasattr(line,op):
                yield getattr(line,op)(*args, **kwargs)
            else:
                extext = extend_type(text)
                if hasattr(extext,op):
                    yield getattr(extext,op)(*args, **kwargs)
                else:
                    raise TextOpException('Unknown OP "%s"' % op)

    def __repr__(self):
        rops = []
        for op,args,kwargs in self.ops:
            opargs = [ '%r' % v if isinstance(v,(str,int,float)) else v for v in args ]
            opargs += [ '%s=%r' % (k,v) for k,v in list(kwargs.items()) ]
            rops.append('%s(%s)' % (op,','.join(map(str,opargs))))
        return '.'.join(rops)

    @classmethod
    def make_gen(cls, text, return_if_none=None):
        if text in [None,NoAttr]:
            return return_if_none
        return cls._tolist(text)

    @property
    def g(self):
        r"""Execute operations, return a generator when possible or a list otherwise

        This is to be used ONLY when the input text has be set as the first argument of the first
        operation.

        Examples:

            >>> echo('hello')
            echo('hello')
            >>> echo('hello').g
            ['hello']
            >>> def mygen(): yield 'hello'
            >>> cut(mygen(),'l')                                # doctest: +ELLIPSIS
            cut(<generator object mygen at ...>,'l')
            >>> cut(mygen(),'l').g                              # doctest: +ELLIPSIS
            <generator object extend_type_gen at ...>
            >>> def mygen(): yield None
            >>> type(echo(None).g)                              # doctest: +ELLIPSIS
            <class 'NoneType'>
        """
        text = self._process()
        return self.make_gen(text)

    @property
    def ge(self):
        r"""Execute operations, return a generator when possible or a list otherwise,
        ( [] if the result is None ).

        This works like :attr:`g` except it returns an empty list if the execution
        result is None.

        Examples:

            >>> echo(None).ge                                    # doctest: +ELLIPSIS
            []
        """
        text = self._process()
        return self.make_gen(text,[])

    @classmethod
    def make_list(cls, text, return_if_none=None):
        if text in [None,NoAttr]:
            return return_if_none
        elif isinstance(text, str):
            return text.splitlines()
        elif isinstance(text, (types.GeneratorType, enumerate)):
            return ListExt(text)
        elif isinstance(text, (int, float)):
            return ListExt([text])
        return text

    @property
    def l(self):
        r"""Execute operations, return a list

        This is to be used ONLY when the input text has be set as the first argument of the first
        operation.

        Examples:

            >>> echo('hello')
            echo('hello')
            >>> echo('hello').l
            ['hello']
            >>> type(echo(None).g)
            <class 'NoneType'>
        """
        text = self._process()
        return self.make_list(text)

    @property
    def le(self):
        r"""Execute operations, returns a list ( [] if the result is None ).

        This works like :attr:`l` except it returns an empty list if the execution
        result is None.

        Examples:

            >>> echo(None).le
            []
        """
        text = self._process()
        return self.make_list(text,[])

    @classmethod
    def make_string(cls, text, join_str='\n', return_if_none=None):
        if text in [None,NoAttr]:
            return return_if_none
        elif isinstance(text, (list,types.GeneratorType,abc.ItemsView,abc.KeysView,abc.ValuesView)):
            return StrExt(join_str.join([ stru(item) for item in text ]))
        return StrExt(text)

    @property
    def s(self):
        r"""Execute operations, return a string (join = newline)

        This is to be used ONLY when the input text has be set as the first argument of the first
        operation. If the result is a list or a generator, it is converted into a string by joinning
        items with a newline.

        Examples:

            >>> echo('hello')
            echo('hello')
            >>> echo('hello').s
            'hello'
            >>> echo(['hello','world']).s
            'hello\nworld'
            >>> type(echo(None).s)
            <class 'NoneType'>
        """
        text = self._process()
        return self.make_string(text)

    @property
    def se(self):
        r"""Execute operations, returns a string ( '' if the result is None ).

        This works like :attr:`s` except it returns an empty string if the execution
        result is None.

        Examples:

            >>> echo(None).se
            ''
        """
        text = self._process()
        return self.make_string(text,return_if_none='')

    @property
    def j(self):
        r"""Execute operations, return a string (join = '')

        This works like :attr:`s` except that joins will be done with an empty string

        Examples:

            >>> echo(['hello','world']).j
            'helloworld'
            >>> type(echo(None).j)
            <class 'NoneType'>
        """
        text = self._process()
        return self.make_string(text,join_str='')

    @property
    def je(self):
        r"""Execute operations, returns a string ( '' if the result is None, join='').

        This works like :attr:`j` except it returns an empty string if the execution
        result is None.

        Examples:

            >>> echo(None).je
            ''
        """
        text = self._process()
        return self.make_string(text,join_str='',return_if_none='')

    @classmethod
    def make_int(cls, text):
        def _to_int(text):
            try:
                return int(float(text))
            except (ValueError, TypeError):
                return 0
        if isinstance(text, str):
            return _to_int(text)
        elif isinstance(text,(list,tuple)):
            return list(map(_to_int,text))
        return text

    @property
    def i(self):
        r"""Execute operations, returns an int.

        Examples:

            >>> echo('1789').i
            1789
            >>> echo('3.14').i
            3
            >>> echo('Tea for 2').i
            0
        """
        text = self._process()
        return self.make_int(text)

    @classmethod
    def make_float(cls, text):
        def _to_float(text):
            try:
                return float(text)
            except (ValueError, TypeError):
                return 0.0
        if isinstance(text, str):
            return _to_float(text)
        elif isinstance(text, (list,tuple)):
            return list(map(_to_float,text))
        return text

    @property
    def f(self):
        r"""Execute operations, returns a float.

        Examples:

            >>> echo('1789').f
            1789.0
            >>> echo('3.14').f
            3.14
            >>> echo('Tea for 2').f
            0.0
        """
        text = self._process()
        return self.make_float(text)

    @property
    def r(self):
        r"""Execute operations, do not convert.

        Examples:

            >>> echo('1789').length().l
            [4]
            >>> echo('1789').length().s
            '4'
            >>> echo('1789').length().r
            4
        """
        return self._process()

    @classmethod
    def consume(cls, text):
        if isinstance(text, types.GeneratorType):
            for i in text:
                pass

    @property
    def n(self):
        r"""Execute operations, do not convert, do not return anything

        If _process() returns a generator, it is consumed

        Examples:

            >>> echo('1789').length().n

        """
        text = self._process()
        self.consume(text)

    @property
    def pp(self):
        r"""Execute operations, return Prettyprint version of the result

        Examples:

        >>> s = '''
        ... a:val1
        ... b:
        ...     c:val3
        ...     d:
        ...         e ... : val5
        ...         f ... :val6
        ...     g:val7
        ... f: val8'''
        >>> print(parse_indented(s).r)
        {'a': 'val1', 'b': {'c': 'val3', 'd': {'e': 'val5', 'f': 'val6'}, 'g': 'val7'}, 'f': 'val8'}
        >>> print(parse_indented(s).pp)
        {   'a': 'val1',
            'b': {'c': 'val3', 'd': {'e': 'val5', 'f': 'val6'}, 'g': 'val7'},
            'f': 'val8'}
        """
        text = self._process()
        return pp.pformat(text)

    @classmethod
    def op(cls,text,*args,**kwargs):
        """ This method must be overriden in derived classes """
        return cls.fn(text,*args,**kwargs)

    @classmethod
    def _tolist(cls,text):
        if isinstance(text, str):
            return str.splitlines(text)
        elif isinstance(text, bytes):
            return bytes.splitlines(text)
        elif isinstance(text, dict):
            return iter(text.items())
        elif isinstance(text, (int,float)):
            return [text]
        return text

    @classmethod
    def _tosublist(cls,text):
        if isinstance(text, str):
            return [ [line] for line in str.splitlines(text) ]
        elif isinstance(text, bytes):
            return [ [line] for line in bytes.splitlines(text) ]
        elif isinstance(text, (int,float,dict)):
            return [[text]]
        elif isinstance(text, list) and (not text or not isinstance(text[0],list)):
            return [text]
        return text

    @classmethod
    def _tostr(cls,text):
        if not isinstance(text, str):
            return '\n'.join(text)
        return text

    @classmethod
    def make_dict(cls, text):
        return DictExt(text)

    def __len__(self):
        text = self._process()
        if isinstance(text, types.GeneratorType):
            return sum(1 for x in text)
        return len(text)

    def __getitem__(self, item):
        lst = self.l
        return lst.__getitem__(item)

def extend_type(obj):
    if isinstance(obj,bytes):
        if not isinstance(obj,BytesExt):
            return BytesExt(obj)
    elif isinstance(obj,str):
        if not isinstance(obj,StrExt):
            return StrExt(obj)
    elif isinstance(obj, tuple):
        if not isinstance(obj, TupleExt):
            return TupleExt(obj)  # mandatory to have pickle working on DictExt
    elif isinstance(obj,(list,abc.ItemsView,abc.KeysView,abc.ValuesView)):
        if not isinstance(obj,ListExt):
            return ListExt(obj)
    elif isinstance(obj,dict):
        if not isinstance(obj,DictExt):
            return DictExt(obj)
    elif isinstance(obj,types.GeneratorType):
        return extend_type_gen(obj)
    return obj

def extend_type_gen(obj):
    for i in obj:
        yield extend_type(i)

def stru(text):
    if isinstance(text, bytes):
        text = text.decode('utf-8','replace')
    else:
        text = str(text)
    return text

def set_debug(flag):
    """ Change debug level

    It sets logger level to DEBUG if flag is True, otherwise to CRITICAL (no log at all)
    """
    logger.setLevel(flag and logging.DEBUG or logging.CRITICAL)

class WrapOp(TextOp):
    input_argn = 0
    # fn=<to be defined in child class>
    @classmethod
    def op(cls, text, *args,**kwargs):
        args = list(args)
        args.insert(cls.input_argn,text)
        return cls.fn(*args,**kwargs)

class WrapOpStr(TextOp):
    input_argn = 0
    # fn=<to be defined in child class>
    @classmethod
    def op(cls, text, *args,**kwargs):
        args = list(args)
        if isinstance(text, str):
            args.insert(cls.input_argn,text)
            return cls.fn(*args,**kwargs)
        else:
            return wrap_op_str_gen(text, cls.fn, cls.input_argn, args, kwargs)

def wrap_op_str_gen(text, fn, argn, args, kwargs):
    args.insert(argn,None)
    for line in text:
        args[argn] = line
        yield fn(*args,**kwargs)

class WrapOpIter(TextOp):
    input_argn = 0
    # fn=<to be defined in child class>
    @classmethod
    def op(cls, text,*args,**kwargs):
        args = list(args)
        args.insert(cls.input_argn,cls._tolist(text))
        return cls.fn(*args,**kwargs)

def add_textop(class_or_func):
    """Decorator to declare custom function or custom class as a new textops op

    the custom function/class will receive the whole raw input text at once.

    Examples:

        >>> @add_textop
        ... def repeat(text, n, *args,**kwargs):
        ...     return text * n
        >>> 'hello' | repeat(3)
        'hellohellohello'

        >>> @add_textop
        ... class cool(TextOp):
        ...     @classmethod
        ...     def op(cls, text, *args,**kwargs):
        ...         return text + ' is cool.'
        >>> 'textops' | cool()
        'textops is cool.'
    """
    if isinstance(class_or_func,type):
        op = class_or_func
    else:
        op = type(class_or_func.__name__,(TextOp,), {'fn':staticmethod(class_or_func)})

    setattr(textops.ops,class_or_func.__name__,op)
    return op

def add_textop_iter(func):
    r"""Decorator to declare custom *ITER* function as a new textops op

    An *ITER* function is a function that will receive the input text as a *LIST* of lines.
    One have to iterate over this list and generate a result (it can be a list, a generator,
    a dict, a string, an int ...)

    Examples:

        >>> @add_textop_iter
        ... def odd(lines, *args,**kwargs):
        ...     for i,line in enumerate(lines):
        ...         if not i % 2:
        ...             yield line
        >>> s = '''line 1
        ... line 2
        ... line 3'''
        >>> s >> odd()
        ['line 1', 'line 3']
        >>> s | odd().tolist()
        ['line 1', 'line 3']

        >>> @add_textop_iter
        ... def sumsize(lines, *args,**kwargs):
        ...     sum = 0
        ...     for line in lines:
        ...         sum += int(re.search(r'\d+',line).group(0))
        ...     return sum
        >>> '''1492 file1
        ... 1789 file2
        ... 2015 file3''' | sumsize()
        5296
    """
    op = type(func.__name__,(WrapOpIter,), {'fn':staticmethod(func)})
    setattr(textops.ops,func.__name__,op)
    return op

class DebugText(object):
    def __init__(self,text,nblines=20,more_msg='...'):
        self.text = text
        self.nblines = nblines
        self.more_msg = more_msg
    def __repr__(self):
        if isinstance(self.text, str):
            lines = str.splitlines(self.text)
            begin,sep,end = '','',''
        elif isinstance(self.text, str):
            lines = str.splitlines(self.text)
            begin,sep,end = '','',''
        elif isinstance(self.text, dict):
            lines = [ '%s : %s' % item for item in self.text.items() ]
            begin,sep,end = '{',',','}'
        elif isinstance(self.text,(types.GeneratorType,list)):
            lines = self.text
            begin,sep,end = '[',',',']'
        else:
            lines = [repr(self.text)]
            begin,sep,end = '','',''

        out = begin
        for i,line in enumerate(lines):
            if i:
                out += '\n'
            if i == self.nblines:
                out += '%s\n' % self.more_msg
                break
            out += '%s%s' % (line[:120],sep)
        out += end
        return out

def get_attribute_or_textop(obj,name):
    if name.startswith('__'):
        return object.__getattribute__(obj,name)
    op_cls = getattr(textops.ops,name,None)
    if op_cls and isinstance(op_cls,type) and issubclass(op_cls,TextOp):
        def fn(*args,**kwargs):
            debug = kwargs.get('debug',False)
            if debug:
                logger.debug('='*60)
                logger.debug('Before %s(%s,%s):', name,args, kwargs)
                logger.debug('%s', DebugText(obj))
            result = op_cls.op(obj,*args,**kwargs)
            if debug:
                logger.debug('-'*60)
                logger.debug('After %s(%s,%s):', name,args, kwargs)
                logger.debug('%s', DebugText(result))
            return result
    else:
        try:
            fn = object.__getattribute__(obj,name)
        except AttributeError:
            if isinstance(obj,DictExt):
                return NoAttr
            elif isinstance(obj,(StrExt,BytesExt)):
                raise
            def fn(*args,**kwargs):
                return [getattr(str if isinstance(s,str) else str, name)(s,*args,**kwargs) for s in obj]

    if not isinstance(fn, abc.Callable):
        return fn

    def wrapper(*args, **kwargs):
        try:
            result = fn(*args, **kwargs)
        except:
            logger.error('*** Error when calling %s(%s,%s)' % (fn.__name__,args, kwargs))
            raise
        if isinstance(result,types.GeneratorType):
            result = list(result)
        return extend_type(result)
    return wrapper

class BytesExt(bytes):
    """Extend bytes class to gain access to textops as attributes

    Examples:

        >>> b'normal bytes'.cut()
        Traceback (most recent call last):
            ...
        AttributeError: 'bytes' object has no attribute 'cut'

        >>> BytesExt(b'extended bytes').cut()
        [b'extended', b'bytes']
    """
    def __getattribute__(self, name):
        return get_attribute_or_textop(self,name)
    @property
    def as_list(self):
        """ Convert to ListExt object """
        return ListExt([self])
    def __getslice__(self,*args, **kwargs):
        return extend_type(super(BytesExt, self).__getslice__(*args, **kwargs))
    def __getitem__(self,*args, **kwargs):
        return extend_type(super(BytesExt, self).__getitem__(*args, **kwargs))
    def __add__(self,*args, **kwargs):
        return extend_type(super(BytesExt, self).__add__(*args, **kwargs))
    def __mul__(self,*args, **kwargs):
        return extend_type(super(BytesExt, self).__mul__(*args, **kwargs))
    def __rmul__(self,*args, **kwargs):
        return extend_type(super(BytesExt, self).__rmul__(*args, **kwargs))
    def __mod__(self,*args, **kwargs):
        return extend_type(super(BytesExt, self).__mod__(*args, **kwargs))
    def __rmod__(self,*args, **kwargs):
        return extend_type(super(BytesExt, self).__rmod__(*args, **kwargs))
    def __format__(self,*args, **kwargs):
        return extend_type(super(BytesExt, self).__format__(*args, **kwargs))
    def _ipython_display_(self):
        print(self.__repr__())

class StrExt(str):
    """Extend str class to gain access to textops as attributes

    Examples:

        >>> 'normal string'.cut()
        Traceback (most recent call last):
            ...
        AttributeError: 'str' object has no attribute 'cut'

        >>> StrExt('extended string').cut()
        ['extended', 'string']
    """
    def __getattribute__(self, name):
        return get_attribute_or_textop(self,name)
    @property
    def as_list(self):
        """ Convert to ListExt object """
        return ListExt([self])
    def __getslice__(self,*args, **kwargs):
        return extend_type(super(StrExt, self).__getslice__(*args, **kwargs))
    def __getitem__(self,*args, **kwargs):
        return extend_type(super(StrExt, self).__getitem__(*args, **kwargs))
    def __add__(self,*args, **kwargs):
        return extend_type(super(StrExt, self).__add__(*args, **kwargs))
    def __mul__(self,*args, **kwargs):
        return extend_type(super(StrExt, self).__mul__(*args, **kwargs))
    def __rmul__(self,*args, **kwargs):
        return extend_type(super(StrExt, self).__rmul__(*args, **kwargs))
    def __mod__(self,*args, **kwargs):
        return extend_type(super(StrExt, self).__mod__(*args, **kwargs))
    def __rmod__(self,*args, **kwargs):
        return extend_type(super(StrExt, self).__rmod__(*args, **kwargs))
    def __format__(self,*args, **kwargs):
        return extend_type(super(StrExt, self).__format__(*args, **kwargs))
    def _ipython_display_(self):
        print(self.__repr__())


class TupleExt(tuple):
    """Extend tuple class to gain access to textops as attributes

    In addition, all tuple items are extended on-the-fly when accessed

    Examples:

        >>> ('normal','list').grep('t')
        Traceback (most recent call last):
            ...
        AttributeError: 'tuple' object has no attribute 'grep'

        >>> TupleExt(('extended','list')).grep('t')
        ['extended', 'list']

        Note: grep() returns a list
    """
    def __getattribute__(self, name):
        return get_attribute_or_textop(self,name)
    @property
    def as_tuple(self):
        """ Convert to TupleExt object """
        return self
    def __getslice__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__getslice__(*args, **kwargs))
    def __getitem__(self,key,*args, **kwargs):
        try:
            return extend_type(super().__getitem__(key,*args, **kwargs))
        except IndexError:
            return NoAttr
    def __add__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__add__(*args, **kwargs))
    def __mul__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__mul__(*args, **kwargs))
    def __iadd__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__iadd__(*args, **kwargs))
    def __imul__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__imul__(*args, **kwargs))
    def __rmul__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__rmul__(*args, **kwargs))
    def __format__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__format__(*args, **kwargs))
    def __iter__(self):
        return TupleExtIterator(self)
    def _ipython_display_(self):
        print(self.__repr__())

class TupleExtIterator(object):
    def __init__(self, obj):
        self.obj = obj
        self.index = -1
        self.len = len(obj)

    def __iter__(self):
        return self

    def __next__(self):
        self.index += 1
        if self.index < self.len:
            # Will call extend_type() because __getitem__ do it :
            return self.obj[self.index]
        else:
            raise StopIteration


class ListExt(list):
    """Extend list class to gain access to textops as attributes

    In addition, all list items (dict, list, str, bytes) are extended on-the-fly when accessed

    Examples:

        >>> ['normal','list'].grep('t')
        Traceback (most recent call last):
            ...
        AttributeError: 'list' object has no attribute 'grep'

        >>> ListExt(['extended','list']).grep('t')
        ['extended', 'list']
    """
    def __getattribute__(self, name):
        return get_attribute_or_textop(self,name)
    @property
    def as_list(self):
        """ Convert to ListExt object """
        return self
    def __getslice__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__getslice__(*args, **kwargs))
    def __getitem__(self,key,*args, **kwargs):
        try:
            return extend_type(super(ListExt, self).__getitem__(key,*args, **kwargs))
        except IndexError:
            return NoAttr
    def __add__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__add__(*args, **kwargs))
    def __mul__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__mul__(*args, **kwargs))
    def __iadd__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__iadd__(*args, **kwargs))
    def __imul__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__imul__(*args, **kwargs))
    def __rmul__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__rmul__(*args, **kwargs))
    def __format__(self,*args, **kwargs):
        return extend_type(super(ListExt, self).__format__(*args, **kwargs))
    def __iter__(self):
        return ListExtIterator(self)
    def _ipython_display_(self):
        print(self.__repr__())

class ListExtIterator(object):
    def __init__(self, obj):
        self.obj = obj
        self.index = -1
        self.len = len(obj)
    def __iter__(self):
        return self
    def __next__(self):
        self.index += 1
        if self.index < self.len:
            # Will call extend_type() because __getitem__ do it :
            return self.obj[self.index]
        else:
            raise StopIteration

class DictExt(NoAttrDict):
    """Extend dict class with new features

    New features are :

        * Access to textops operations with attribute notation
        * All dict values (dict, list, str, bytes) are extended on-the-fly when accessed
        * Access to dict values with attribute notation
        * Add a key:value in the dict with attribute notation (one level at a time)
        * Returns NoAttr object when a key is not in the Dict
        * add modification on-the-fly :meth:`amend` and rendering to string :meth:`render`

    Note:

        ``NoAttr`` is a special object that returns always ``NoAttr`` when accessing to any attribute.
        it behaves like ``False`` for testing, ``[]`` in foor-loops. The goal is to be able
        to use very long expression with dotted notation without being afraid to get an exception.

    Examples:

        >>> {'a':1,'b':2}.items().grep('a')
        Traceback (most recent call last):
            ...
        AttributeError: 'dict_items' object has no attribute 'grep'

        >>> DictExt({'a':1,'b':2}).items().grep('a')
        [('a', 1)]

        >>> d = DictExt({ 'this' : { 'is' : { 'a' : {'very deep' : { 'dict' : 'yes it is'}}}}})
        >>> print(d.this['is'].a['very deep'].dict)
        yes it is
        >>> d.not_a_valid_key
        NoAttr
        >>> d['not_a_valid_key']
        NoAttr
        >>> d.not_a_valid_key.and_i.can.put.things.after.without.exception
        NoAttr
        >>> for obj in d.not_a_valid_key.objects:
        ...     do_things(obj)
        ... else:
        ...     print('no object')
        no object

        >>> d = DictExt()
        >>> d.a = DictExt()
        >>> d.a.b = 'this is my logging data'
        >>> print(d)
        {'a': {'b': 'this is my logging data'}}

        >>> d = { 'mykey' : 'myval' }
        >>> d['mykey']
        'myval'
        >>> type(d['mykey'])
        <class 'str'>
        >>> d = DictExt(d)
        >>> d['mykey']
        'myval'
        >>> type(d['mykey'])
        <class 'textops.base.StrExt'>

        >>> d=DictExt()
        >>> d[0]=[]
        >>> d
        {0: []}
        >>> d[0].append(3)
        >>> d
        {0: [3]}
        >>> type(d[0])
        <class 'textops.base.ListExt'>

    """
    def __getattribute__(self, name):
        if dict.__contains__(self,name):
            return dict.__getitem__(self,name)
        return get_attribute_or_textop(self,name)

    @property
    def as_list(self):
        """ Convert to ListExt object """
        return ListExt([self])

    def amend(self,*args, **kwargs):
        """ Modify on-the-fly a dictionary

        The method will generate a new extended dictionary and update it with given params

        Examples:

            >>> s = '''soft:textops
            ... count:32591'''
            >>> s | parse_indented()
            {'soft': 'textops', 'count': '32591'}
            >>> s | parse_indented().amend(date='2015-11-19')
            {'soft': 'textops', 'count': '32591', 'date': '2015-11-19'}
        """
        return DictExt(self,*args, **kwargs)
    def render(self,format_string,defvalue='-'):
        """ Render a DictExt as a string

        It uses the fonction :func:`dformat` to format the dictionary

        Args:
            format_string (str): Same format string as for :meth:`str.format`
            defvalue (str or callable): the default value to display when the data is not in the dict

        Examples:

            >>> d = DictExt({'count': '32591', 'date': '2015-11-19', 'soft': 'textops'})
            >>> d.render('On {date}, "{soft}" has been downloaded {count} times')
            'On 2015-11-19, "textops" has been downloaded 32591 times'
            >>> d.render('On {date}, "{not_in_dict}" has been downloaded {count} times','?')
            'On 2015-11-19, "?" has been downloaded 32591 times'
        """
        return dformat(format_string,self,defvalue)
    def __getitem__(self,*args, **kwargs):
        return extend_type(super(DictExt, self).__getitem__(*args, **kwargs))
    def __format__(self,*args, **kwargs):
        return extend_type(super(DictExt, self).__format__(*args, **kwargs))

class DefaultDict(dict):
    def __init__(self,defvalue,*args,**kwargs):
        self.defvalue = defvalue
        super(DefaultDict,self).__init__(*args,**kwargs)
    def __getitem__(self,key):
        try:
            return super(DefaultDict,self).__getitem__(key)
        except KeyError:
            if isinstance(self.defvalue, abc.Callable):
                return self.defvalue(key)
            return self.defvalue

class DefaultList(list):
    def __init__(self,defvalue,*args,**kwargs):
        self.defvalue = defvalue
        super(DefaultList,self).__init__(*args,**kwargs)
    def __getitem__(self,key):
        try:
            return super(DefaultList,self).__getitem__(key)
        except IndexError:
            if isinstance(self.defvalue, abc.Callable):
                return self.defvalue(key)
            return self.defvalue

def dictmerge(*dict_args):
    """Merge as many dicts you want

    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.

    Args:
        *dict_args (dict): List of dicts

    Returns:
        dict: a new merged dict

    Examples:

        >>> dictmerge({'a':1,'b':2},{'b':3,'c':4})
        {'a': 1, 'b': 3, 'c': 4}

    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result

string_formatter = string.Formatter()
vformat = string_formatter.vformat

def dformat(format_str,dct,defvalue='-'):
    """ Formats a dictionary, manages unkown keys

    It works like :meth:`string.Formatter.vformat` except that it accepts only a dict
    for values and a defvalue for not matching keys. Defvalue can be a callable that will
    receive the requested key as argument and return a string

    Args:
        format_string (str): Same format string as for :meth:`str.format`
        dct (dict) : the dict to format
        defvalue (str or callable): the default value to display when the data is not in the dict

    Examples:

        >>> d = {'count': '32591', 'soft': 'textops'}
        >>> dformat('{soft} : {count} dowloads',d)
        'textops : 32591 dowloads'
        >>> dformat('{software} : {count} dowloads',d,'N/A')
        'N/A : 32591 dowloads'
        >>> dformat('{software} : {count} dowloads',d,lambda k:'unknown_tag_%s' % k)
        'unknown_tag_software : 32591 dowloads'
    """
    return vformat(format_str,(),DefaultDict(defvalue,dct))

def eformat(format_str,lst,dct,defvalue='-'):
    """ Formats a list and a dictionary, manages unkown keys

    It works like :meth:`string.Formatter.vformat` except that it accepts a defvalue for not matching keys.
    Defvalue can be a callable that will receive the requested key as argument and return a string

    Args:
        format_string (str): Same format string as for :meth:`str.format`
        lst (dict) : the list to format
        dct (dict) : the dict to format
        defvalue (str or callable): the default value to display when the data is not in the dict

    Examples:

        >>> d = {'count': '32591', 'soft': 'textops'}
        >>> l = ['Eric','Guido']
        >>> eformat('{0} => {soft} : {count} dowloads',l,d)
        'Eric => textops : 32591 dowloads'
        >>> eformat('{2} => {software} : {count} dowloads',l,d,'N/A')
        'N/A => N/A : 32591 dowloads'
        >>> eformat('{2} => {software} : {count} dowloads',l,d,lambda k:'unknown_tag_%s' % k)
        'unknown_tag_2 => unknown_tag_software : 32591 dowloads'
    """
    return vformat(format_str,DefaultList(defvalue,lst),DefaultDict(defvalue,dct))


def decode_bytes(byte_string, encoding=None):
    """Returns a decoded string from a byte string.
    It can test several encodings, will try to detect encoding if it does not
    work and have a ascii/'replace' fallback.
    One can give standard python3 string, in which case, no action is done

    Args:

        byte_string (bytes): string to decode to python3 standard string
        encoding (str, list, tuple): ending(s) to use. if a list/tuple is given,
            will use the first that does not raise exception
            default : 'utf-8'

    Examples:

        >>> b = 'éric'.encode('utf-16')
        >>> print(b.hex())
        fffee900720069006300
        >>> decode_bytes(b)
        'éric'
        >>> decode_bytes(b,'utf-16')
        'éric'
        >>> decode_bytes(b,('utf-16','utf-32'))
        'éric'

    Returns:

        a python3 standard string
    """
    if not isinstance(byte_string, bytes):
        return byte_string
    if not encoding:
        encoding = ('utf-8',)
    if isinstance(encoding, str):
        encoding = (encoding,)
    for enc in encoding:
        try:
            s = byte_string.decode(enc)
            return s
        except UnicodeDecodeError:
            pass
    # If given encoding does not work try to detect it :
    detected_encoding = chardet.detect(byte_string)
    if detected_encoding['confidence'] >= 0.5:
        try:
            s = byte_string.decode(detected_encoding['encoding'])
            return s
        except UnicodeDecodeError:
            pass
    # Fallback : ascii + replace unkown chars
    s = byte_string.decode('ascii','replace')
    return s


# NoAttr.as_list must return an empty ListExt() not a simple list
# This customize noattr module for textops.
from noattr import NoAttrType
def _as_list(self):
    return ListExt()
NoAttrType.as_list=property(_as_list)



# -*- coding: utf-8 -*-
#
# Created : 2015-08-24
#
# @author: Eric Lapouyade
#
""" This module gathers parsers to handle whole input text"""

from textops import TextOp, NoAttr, dformat, pp, stru
import textops
import types
import string
import re
import copy
from datetime import datetime
import collections

logger = textops.logger

def index_normalize(index_val):
    """Normalize dictionary calculated key

    When parsing, keys within a dictionary may come from the input text. To ensure there is no
    space or other special caracters, one should use this function. This is useful because
    DictExt dictionaries can be access with a dotted notation that only supports ``A-Za-z0-9_`` chars.

    Args:
        index_val (str): The candidate string to a dictionary key.

    Returns:
        str: A normalized string with only ``A-Za-z0-9_`` chars

    Examples:
        >>> index_normalize('this my key')
        'this_my_key'
        >>> index_normalize('this -my- %key%')
        'this_my_key'

    """
    index_val = index_val.lower().strip()
    index_val = re.sub(r'^\W*','',index_val)
    index_val = re.sub(r'\W*$','',index_val)
    index_val = re.sub(r'\W+','_',index_val)
    index_val = re.sub('_+','_',index_val)
    return index_val

def context_key_not_found(key):
    return 'UNKNOWN_CONTEXT_KEY_%s' % key

class ParsingError(Exception):
    pass

class mgrep(TextOp):
    r"""Multiple grep

    This works like :class:`textops.grep` except that it can do several greps in a single command.
    By this way, you can select many patterns in a big file.

    Args:
        patterns_dict (dict): a dictionary where all patterns to search are in values.
        key (int or str): test only one column or one key (optional)

    Returns:
        dict: A dictionary where the keys are the same as for ``patterns_dict``, the values will
            contain the :class:`textops.grep` result for each corresponding patterns.

    Examples:
        >>> logs = '''
        ... error 1
        ... warning 1
        ... warning 2
        ... info 1
        ... error 2
        ... info 2
        ... '''
        >>> t = logs | mgrep({
        ... 'errors' : r'^err',
        ... 'warnings' : r'^warn',
        ... 'infos' : r'^info',
        ... })
        >>> print(t)  #doctest: +NORMALIZE_WHITESPACE
        {'errors': ['error 1', 'error 2'],
        'warnings': ['warning 1', 'warning 2'],
        'infos': ['info 1', 'info 2']}

        >>> s = '''
        ... Disk states
        ... -----------
        ... name: c1t0d0s0
        ... state: good
        ... fs: /
        ... name: c1t0d0s4
        ... state: failed
        ... fs: /home
        ...
        ... '''
        >>> t = s | mgrep({
        ... 'disks' : r'^name:',
        ... 'states' : r'^state:',
        ... 'fss' : r'^fs:',
        ... })
        >>> print(t)  #doctest: +NORMALIZE_WHITESPACE
        {'disks': ['name: c1t0d0s0', 'name: c1t0d0s4'],
        'states': ['state: good', 'state: failed'],
        'fss': ['fs: /', 'fs: /home']}
        >>> dict(zip(t.disks.cutre(': *',1),zip(t.states.cutre(': *',1),t.fss.cutre(': *',1))))
        {'c1t0d0s0': ('good', '/'), 'c1t0d0s4': ('failed', '/home')}
    """
    flags = 0
    reverse = False
    @classmethod
    def op(cls,text,patterns_dict,key = None, *args,**kwargs):
        for k,pattern in list(patterns_dict.items()):
            if isinstance(pattern,str):
                patterns_dict[k] = re.compile(pattern,cls.flags)
        dct = {}
        for line in cls._tolist(text):
            for k,regex in list(patterns_dict.items()):
                try:
                    if isinstance(line,str):
                        if bool(regex.search(stru(line))) != cls.reverse:  # kind of XOR with cls.reverse
                            dct.setdefault(k,[]).append(line)
                    elif key is None:
                        if bool(regex.search(stru(line))) != cls.reverse:  # kind of XOR with cls.reverse
                            dct.setdefault(k,[]).append(line)
                    else:
                        if bool(regex.search(stru(line[key]))) != cls.reverse:  # kind of XOR with cls.reverse
                            dct.setdefault(k,[]).append(line)
                except (ValueError, TypeError, IndexError, KeyError):
                    pass
        return dct

class mgrepi(mgrep):
    r"""same as mgrep but case insensitive

    This works like :class:`textops.mgrep`, except it is case insensitive.

    Args:
        patterns_dict (dict): a dictionary where all patterns to search are in values.
        key (int or str): test only one column or one key (optional)

    Returns:
        dict: A dictionary where the keys are the same as for ``patterns_dict``, the values will
            contain the :class:`textops.grepi` result for each corresponding patterns.

    Examples:
        >>> 'error 1' | mgrep({'errors':'ERROR'})
        {}
        >>> 'error 1' | mgrepi({'errors':'ERROR'})
        {'errors': ['error 1']}
    """
    flags = re.IGNORECASE

class mgrepv(mgrep):
    r"""Same as mgrep but exclusive

    This works like :class:`textops.mgrep`, except it searches lines that DOES NOT match patterns.

    Args:
        patterns_dict (dict): a dictionary where all patterns to exclude are in values().
        key (int or str): test only one column or one key (optional)

    Returns:
        dict: A dictionary where the keys are the same as for ``patterns_dict``, the values will
            contain the :class:`textops.grepv` result for each corresponding patterns.

    Examples:
        >>> logs = '''error 1
        ... warning 1
        ... warning 2
        ... error 2
        ... '''
        >>> t = logs | mgrepv({
        ... 'not_errors' : r'^err',
        ... 'not_warnings' : r'^warn',
        ... })
        >>> print(t                                         )#doctest: +NORMALIZE_WHITESPACE
        {'not_warnings': ['error 1', 'error 2'], 'not_errors': ['warning 1', 'warning 2']}
    """
    reverse = True

class mgrepvi(mgrepv):
    r"""Same as mgrepv but case insensitive

    This works like :class:`textops.mgrepv`, except it is case insensitive.

    Args:
        patterns_dict (dict): a dictionary where all patterns to exclude are in values().
        key (int or str): test only one column or one key (optional)

    Returns:
        dict: A dictionary where the keys are the same as for ``patterns_dict``, the values will
            contain the :class:`textops.grepvi` result for each corresponding patterns.

    Examples:
        >>> logs = '''error 1
        ... WARNING 1
        ... warning 2
        ... ERROR 2
        ... '''
        >>> t = logs | mgrepv({
        ... 'not_errors' : r'^err',
        ... 'not_warnings' : r'^warn',
        ... })
        >>> print(t                                         )#doctest: +NORMALIZE_WHITESPACE
        {'not_warnings': ['error 1', 'WARNING 1', 'ERROR 2'],
        'not_errors': ['WARNING 1', 'warning 2', 'ERROR 2']}
        >>> t = logs | mgrepvi({
        ... 'not_errors' : r'^err',
        ... 'not_warnings' : r'^warn',
        ... })
        >>> print(t                                         )#doctest: +NORMALIZE_WHITESPACE
        {'not_warnings': ['error 1', 'ERROR 2'], 'not_errors': ['WARNING 1', 'warning 2']}
    """
    flags = re.IGNORECASE

class sgrep(TextOp):
    r"""Switch grep

    This works like :class:`textops.mgrep` except that it returns a list of lists.
    ``sgrep`` dispatches lines matching a pattern to the list corresponding to the pattern order.
    If a line matches the third pattern, it will be dispatched to the third returned list.
    If N patterns are given to search, it will return N+1 lists, where the last list will be filled
    of lines that does not match any pattern in the given patterns list.
    The patterns list order is important : only the first
    matching pattern will be taken in account.
    One can consider that ``sgrep`` works like a **switch()** :
    it will do for each line a kind of ::

        if pattern1 matches:
            put line in list1
        elif pattern2 matches:
            put line in list2
        elif patternN matches:
            put line in listN
        else:
            put line in listN+1

    Args:
        patterns (list): a list of patterns to search.
        key (int or str): test only one column or one key (optional)

    Returns:
        list: a list of lists (nb patterns + 1)

    Examples:
        >>> logs = '''
        ... error 1
        ... warning 1
        ... warning 2
        ... info 1
        ... error 2
        ... info 2
        ... '''
        >>> t = logs | sgrep(('^err','^warn'))
        >>> print(t                                         )#doctest: +NORMALIZE_WHITESPACE
        [['error 1', 'error 2'], ['warning 1', 'warning 2'], ['', 'info 1', 'info 2']]
    """
    flags = 0
    reverse = False
    @classmethod
    def op(cls,text,patterns,key = None, *args,**kwargs):
        patterns = [ re.compile(pattern,cls.flags) if isinstance(pattern,str) else pattern for pattern in patterns ]
        lst = [ [] for i in range(len(patterns)+1) ] # surtout pas faire [] * (len(patterns)+1)
        for line in cls._tolist(text):
            for i,regex in enumerate(patterns):
                try:
                    if isinstance(line,str):
                        if bool(regex.search(stru(line))) != cls.reverse:  # kind of XOR with cls.reverse
                            lst[i].append(line)
                            break
                    elif key is None:
                        if bool(regex.search(stru(line))) != cls.reverse:  # kind of XOR with cls.reverse
                            lst[i].append(line)
                            break
                    else:
                        if bool(regex.search(stru(line[key]))) != cls.reverse:  # kind of XOR with cls.reverse
                            lst[i].append(line)
                            break
                except (ValueError, TypeError, IndexError, KeyError):
                    pass
            else:
                lst[-1].append(line)
        return lst

class sgrepi(sgrep):
    r"""Switch grep case insensitive

    This works like :class:`textops.sgrep` but is case insensitive
    """
    flags = re.IGNORECASE

class sgrepv(sgrep):
    r"""Switch grep reversed

    This works like :class:`textops.sgrep` except that it tests that patterns DOES NOT match the line.
    """
    reverse = True

class sgrepvi(sgrepv):
    r"""Switch grep reversed case insensitive

    This works like :class:`textops.sgrepv` but is case insensitive
    """
    flags = re.IGNORECASE

class parseg(TextOp):
    r"""Find all occurrences of one pattern, return MatchObject groupdict

    Args:
        pattern (str): a regular expression string (case sensitive)

    Returns:
        list: A list of dictionaries (MatchObject groupdict)

    Examples:
        >>> s = '''name: Lapouyade
        ... first name: Eric
        ... country: France'''
        >>> s | parseg(r'(?P<key>.*):\s*(?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
        [{'key': 'name', 'val': 'Lapouyade'},
        {'key': 'first name', 'val': 'Eric'},
        {'key': 'country', 'val': 'France'}]
    """
    ignore_case = False
    @classmethod
    def op(cls,text, pattern, *args,**kwargs):
        if isinstance(pattern,str):
            pattern = re.compile(pattern, re.I if cls.ignore_case else 0)
        out = []
        for line in cls._tolist(text):
            m = pattern.match(line)
            if m:
                out.append(m.groupdict())
        return out

class parsegi(parseg):
    r"""Same as parseg but case insensitive

    Args:
        pattern (str): a regular expression string (case insensitive)

    Returns:
        list: A list of dictionaries (MatchObject groupdict)

    Examples:
        >>> s = '''Error: System will reboot
        ... Notice: textops rocks
        ... Warning: Python must be used without moderation'''
        >>> s | parsegi(r'(?P<level>error|warning):\s*(?P<msg>.*)')         #doctest: +NORMALIZE_WHITESPACE
        [{'level': 'Error', 'msg': 'System will reboot'},
        {'level': 'Warning', 'msg': 'Python must be used without moderation'}]
    """
    ignore_case = True

class parsek(TextOp):
    r"""Find all occurrences of one pattern, return one Key

    One have to give a pattern with named capturing parenthesis, the function will return a list
    of value corresponding to the specified key. It works a little like :class:`textops.parseg`
    except that it returns from the groupdict, a value for a specified key ('key' be default)

    Args:
        pattern (str): a regular expression string.
        key_name (str): The key to get ('key' by default)
        key_update (callable): function to convert the found value

    Returns:
        list: A list of values corrsponding to `MatchObject groupdict[key]`

    Examples:
        >>> s = '''Error: System will reboot
        ... Notice: textops rocks
        ... Warning: Python must be used without moderation'''
        >>> s | parsek(r'(?P<level>Error|Warning):\s*(?P<msg>.*)','msg')
        ['System will reboot', 'Python must be used without moderation']
    """
    ignore_case = False
    @classmethod
    def op(cls,text, pattern, key_name = 'key', key_update = None, *args,**kwargs):
        if isinstance(pattern,str):
            pattern = re.compile(pattern, re.I if cls.ignore_case else 0)
        out = []
        for line in cls._tolist(text):
            m = pattern.match(line)
            if m:
                dct = m.groupdict()
                key = dct.get(key_name)
                if key:
                    if key_update:
                        key = key_update(key)
                    out.append(key)
        return out

class parseki(parsek):
    r"""Same as parsek but case insensitive

    It works like :class:`textops.parsek` except the pattern is case insensitive.

    Args:
        pattern (str): a regular expression string.
        key_name (str): The key to get ('key' by default)
        key_update (callable): function to convert the found value

    Returns:
        list: A list of values corrsponding to `MatchObject groupdict[key]`

    Examples:
        >>> s = '''Error: System will reboot
        ... Notice: textops rocks
        ... Warning: Python must be used without moderation'''
        >>> s | parsek(r'(?P<level>error|warning):\s*(?P<msg>.*)','msg')
        []
        >>> s | parseki(r'(?P<level>error|warning):\s*(?P<msg>.*)','msg')
        ['System will reboot', 'Python must be used without moderation']
    """
    ignore_case = True

class parsekv(TextOp):
    r"""Find all occurrences of one pattern, returns a dict of groupdicts

    It works a little like :class:`textops.parseg` except that it returns a dict of dicts :
    values are MatchObject groupdicts, keys are a value in the groupdict at a specified key
    (By default : 'key'). Note that calculated keys are normalized (spaces are replaced by
    underscores)

    Args:
        pattern (str): a regular expression string.
        key_name (str): The key name to optain the value that will be the key of the groupdict
            ('key' by default)
        key_update (callable): function to convert/normalize the calculated key.
            If ``None``, the keys is normalized.
            If not ``None`` but not callable ,the key is unchanged.
        val_name (str): instead of storing the groupdict, on can choose to select
            the value at the key ``val_name`. (by default, None : means the whole groupdict)

    Returns:
        dict: A dict of MatchObject groupdicts

    Examples:
        >>> s = '''name: Lapouyade
        ... first name: Eric
        ... country: France'''
        >>> s | parsekv(r'(?P<key>.*):\s*(?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
        {'name': {'key': 'name', 'val': 'Lapouyade'},
        'first_name': {'key': 'first name', 'val': 'Eric'},
        'country': {'key': 'country', 'val': 'France'}}
        >>> s | parsekv(r'(?P<item>.*):\s*(?P<val>.*)','item',str.upper)         #doctest: +NORMALIZE_WHITESPACE
        {'NAME': {'item': 'name', 'val': 'Lapouyade'},
        'FIRST NAME': {'item': 'first name', 'val': 'Eric'},
        'COUNTRY': {'item': 'country', 'val': 'France'}}
        >>> s | parsekv(r'(?P<key>.*):\s*(?P<val>.*)',key_update=0)         #doctest: +NORMALIZE_WHITESPACE
        {'name': {'key': 'name', 'val': 'Lapouyade'},
        'first name': {'key': 'first name', 'val': 'Eric'},
        'country': {'key': 'country', 'val': 'France'}}
        >>> s | parsekv(r'(?P<key>.*):\s*(?P<val>.*)',val_name='val')         #doctest: +NORMALIZE_WHITESPACE
        {'name': 'Lapouyade', 'first_name': 'Eric', 'country': 'France'}
    """
    ignore_case = False
    val_name = None
    @classmethod
    def op(cls,text, pattern, key_name = 'key', key_update = None, val_name = None, *args,**kwargs):
        if val_name is None:
            val_name = cls.val_name
        if isinstance(pattern,str):
            pattern = re.compile(pattern, re.I if cls.ignore_case else 0)

        def _op(text):
            out = {}
            for line in cls._tolist(text):
                m = pattern.match(line)
                if m:
                    dct = m.groupdict()
                    key = dct.get(key_name)
                    if key:
                        if key_update is None:
                            key_norm = index_normalize(key)
                        elif isinstance(key_update, collections.abc.Callable):
                            key_norm = key_update(key)
                        else:
                            key_norm = key
                        out.update({ key_norm : dct if val_name is None else dct[val_name]})
            return out

        if isinstance(text,(list,types.GeneratorType)):
            return [ _op(item) for item in text ]
        return _op(text)

class parsekvi(parsekv):
    r"""Find all occurrences of one pattern (case insensitive), returns a dict of groupdicts

    It works a little like :class:`textops.parsekv` except that the pattern is case insensitive.

    Args:
        pattern (str): a regular expression string (case insensitive).
        key_name (str): The key name to optain the value that will be the key of the groupdict
            ('key' by default)
        key_update (callable): function to convert/normalize the calculated key.
            If ``None``, the keys is normalized.
            If not ``None`` but not callable ,the key is unchanged.
        val_name (str): instead of storing the groupdict, on can choose to select
            the value at the key ``val_name`. (by default, None : means the whole groupdict)

    Returns:
        dict: A dict of MatchObject groupdicts

    Examples:
        >>> s = '''name: Lapouyade
        ... first name: Eric
        ... country: France'''
        >>> s | parsekvi(r'(?P<key>NAME):\s*(?P<val>.*)')
        {'name': {'key': 'name', 'val': 'Lapouyade'}}
    """
    ignore_case = True

class keyval(parsekv):
    r"""Return a dictionnay where keys and values are taken from the pattern specify

    It is a shortcut for :class:`textops.parsekv` with val_name='val'
    The input can be a string or a list of strings.

    Args:
        pattern (str): a regular expression string.
        key_name (str): The key name to optain the value that will be the key of the groupdict
            ('key' by default)
        key_update (callable): function to convert/normalize the calculated key.
            If ``None``, the keys is normalized.
            If not ``None`` but not callable ,the key is unchanged.
        val_name (str): instead of storing the groupdict, on can choose to select
            the value at the key ``val_name`. (by default, None means 'val')

    Returns:
        dict: A dict of key:val from the matched pattern groupdict or a list of dicts if the input is a list of strings

    Examples:
        >>> s = '''name: Lapouyade
        ... first name: Eric
        ... country: France'''
        >>> s | keyval(r'(?P<key>.*):\s*(?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
        {'name': 'Lapouyade', 'first_name': 'Eric', 'country': 'France'}

        >>> s = [ '''name: Lapouyade
        ... first name: Eric ''',
        ... '''name: Python
        ... first name: Guido''' ]
        >>> s | keyval(r'(?P<key>.*):\s*(?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
        [{'name': 'Lapouyade', 'first_name': 'Eric '}, {'name': 'Python', 'first_name': 'Guido'}]

    """
    val_name = 'val'

class keyvali(keyval):
    r"""Return a dictionnay where keys and values are taken from the pattern specify

    It works a little like :class:`textops.keyval` except that the pattern is case insensitive.

    Args:
        pattern (str): a regular expression string (case insensitive).
        key_name (str): The key name to optain the value that will be the key of the groupdict
            ('key' by default)
        key_update (callable): function to convert/normalize the calculated key.
            If ``None``, the keys is normalized.
            If not ``None`` but not callable ,the key is unchanged.
        val_name (str): instead of storing the groupdict, on can choose to select
            the value at the key ``val_name`. (by default, None means 'val')

    Returns:
        dict: A dict of key:val from the matched pattern groupdict

    Examples:
        >>> s = '''name IS Lapouyade
        ... first name IS Eric
        ... country IS France'''
        >>> s | keyvali(r'(?P<key>.*) is (?P<val>.*)')         #doctest: +NORMALIZE_WHITESPACE
        {'name': 'Lapouyade', 'first_name': 'Eric', 'country': 'France'}
    """
    ignore_case = True

class find_pattern(TextOp):
    r"""Fast pattern search

    This operation can be use to find a pattern very fast : it uses :func:`re.search` on the whole input
    text at once. The input text is not read line by line, this means it must fit into memory.
    It returns the first captured group (named or not named group).

    Args:
        pattern (str): a regular expression string (case sensitive).

    Returns:
        str: the first captured group or NoAttr if not found

    Examples:
        >>> s = '''This is data text
        ... Version: 1.2.3
        ... Format: json'''
        >>> s | find_pattern(r'^Version:\s*(.*)')
        '1.2.3'
        >>> s | find_pattern(r'^Format:\s*(?P<format>.*)')
        'json'
        >>> s | find_pattern(r'^version:\s*(.*)') # 'version' : no match because case sensitive
        NoAttr
    """
    ignore_case = False

    @classmethod
    def op(cls,text, pattern, *args,**kwargs):
        if isinstance(pattern,str):
            pattern = re.compile(pattern, re.M | (re.I if cls.ignore_case else 0))
        text = cls._tostr(text)
        m = pattern.search(text)
        if m :
            grps = m.groups()
            return grps[0] if grps else NoAttr
        return NoAttr

class find_patterni(find_pattern):
    r"""Fast pattern search (case insensitive)

    It works like :class:`textops.find_pattern` except that the pattern is case insensitive.

    Args:
        pattern (str): a regular expression string (case insensitive).

    Returns:
        str: the first captured group or NoAttr if not found

    Examples:
        >>> s = '''This is data text
        ... Version: 1.2.3
        ... Format: json'''
        >>> s | find_patterni(r'^version:\s*(.*)')
        '1.2.3'
    """
    ignore_case=True

class find_patterns(TextOp):
    r"""Fast multiple pattern search

    It works like :class:`textops.find_pattern` except that one can specify a list or a dictionary
    of patterns. Patterns must contains capture groups.
    It returns a list or a dictionary of results depending on the patterns argument type.
    Each result will be the re.MatchObject groupdict if there
    are more than one capture named group in the pattern otherwise directly the value corresponding
    to the unique captured group.
    It is recommended to use *named* capture group, if not, the groups will be automatically named
    'groupN' with N the capture group order in the pattern.

    Args:
        patterns (list or dict): a list or a dictionary of patterns.

    Returns:
        dict: patterns search result

    Examples:
        >>> s = '''This is data text
        ... Version: 1.2.3
        ... Format: json'''
        >>> r = s | find_patterns({
        ... 'version':r'^Version:\s*(?P<major>\d+)\.(?P<minor>\d+)\.(?P<build>\d+)',
        ... 'format':r'^Format:\s*(?P<format>.*)',
        ... })
        >>> r
        {'version': {'major': '1', 'minor': '2', 'build': '3'}, 'format': 'json'}
        >>> r.version.major
        '1'
        >>> s | find_patterns({
        ... 'version':r'^Version:\s*(\d+)\.(\d+)\.(\d+)',
        ... 'format':r'^Format:\s*(.*)',
        ... })
        {'version': {'group0': '1', 'group1': '2', 'group2': '3'}, 'format': 'json'}
        >>> s | find_patterns({'version':r'^version:\s*(.*)'}) # lowercase 'version' : no match
        {}
        >>> s = '''creation: 2015-10-14
        ... update: 2015-11-16
        ... access: 2015-11-17'''
        >>> s | find_patterns([r'^update:\s*(.*)', r'^access:\s*(.*)', r'^creation:\s*(.*)'])
        ['2015-11-16', '2015-11-17', '2015-10-14']
        >>> s | find_patterns([r'^update:\s*(?P<year>.*)-(?P<month>.*)-(?P<day>.*)',
        ... r'^access:\s*(.*)', r'^creation:\s*(.*)'])
        [{'year': '2015', 'month': '11', 'day': '16'}, '2015-11-17', '2015-10-14']
    """
    stop_when_found = False
    ignore_case = False

    @classmethod
    def op(cls,text, patterns,*args,**kwargs):
        out = []
        text = cls._tostr(text)
        if isinstance(patterns, dict):
            patterns_list = list(patterns.items())
        else:
            patterns_list = enumerate(patterns)
        for attr,pattern in patterns_list:
            if isinstance(pattern,str):
                pattern = re.compile(pattern, re.M | (re.I if cls.ignore_case else 0))
            if pattern:
                m = pattern.search(text)
                if m :
                    tmp_groupdict = m.groupdict() or dict([('group%s' % k,v) for k,v in enumerate(m.groups())])
                    groupdict = {}
                    for grp, val in list(tmp_groupdict.items()):
                        if grp[:3] == 'INT':
                            try:
                                groupdict[grp[3:]] = int(val)
                            except ValueError:
                                groupdict[grp[3:]] = 0
                        else:
                            groupdict[grp] = val
                    groupdict = cls.pre_store(groupdict)
                    if len(groupdict) == 1:
                        out.append((attr, groupdict.popitem()[1]))
                    else:
                        out.append((attr, groupdict))
                    if cls.stop_when_found:
                        break
        if isinstance(patterns, dict):
            out = dict(out)
        else:
            out = [ i[1] for i in out ]
        return out

    @classmethod
    def pre_store(self,groupdict):
        return groupdict

class find_patternsi(find_patterns):
    r"""Fast multiple pattern search (case insensitive)

    It works like :class:`textops.find_patterns` except that patterns are case insensitive.

    Args:
        patterns (dict): a dictionary of patterns.

    Returns:
        dict: patterns search result

    Examples:
        >>> s = '''This is data text
        ... Version: 1.2.3
        ... Format: json'''
        >>> s | find_patternsi({'version':r'^version:\s*(.*)'})     # case insensitive
        {'version': '1.2.3'}
    """
    ignore_case=True


class find_first_pattern(find_patterns):
    r"""Fast multiple pattern search, returns on first match

    It works like :class:`textops.find_patterns` except that it stops searching on first match.

    Args:
        patterns (list): a list of patterns.

    Returns:
        str or dict: matched value if only one capture group otherwise the full groupdict

    Examples:
        >>> s = '''creation: 2015-10-14
        ... update: 2015-11-16
        ... access: 2015-11-17'''
        >>> s | find_first_pattern([r'^update:\s*(.*)', r'^access:\s*(.*)', r'^creation:\s*(.*)'])
        '2015-11-16'
        >>> s | find_first_pattern([r'^UPDATE:\s*(.*)'])
        NoAttr
        >>> s | find_first_pattern([r'^update:\s*(?P<year>.*)-(?P<month>.*)-(?P<day>.*)'])
        {'year': '2015', 'month': '11', 'day': '16'}
    """
    stop_when_found = True

    @classmethod
    def op(cls,text, patterns,*args,**kwargs):
        data = super(find_first_pattern,cls).op(text, patterns)
        if not data:
            return NoAttr
        return data[0]

class find_first_patterni(find_first_pattern):
    r"""Fast multiple pattern search, returns on first match

    It works like :class:`textops.find_first_pattern` except that patterns are case insensitives.

    Args:
        patterns (list): a list of patterns.

    Returns:
        str or dict: matched value if only one capture group otherwise the full groupdict

    Examples:
        >>> s = '''creation: 2015-10-14
        ... update: 2015-11-16
        ... access: 2015-11-17'''
        >>> s | find_first_patterni([r'^UPDATE:\s*(.*)'])
        '2015-11-16'
    """
    ignore_case=True

class parse_indented(TextOp):
    r"""Parse key:value indented text

    It looks for key:value patterns, store found values in a dictionary. Each time a new indent is
    found, a sub-dictionary is created. The keys are normalized (only keep ``A-Za-z0-9_``), the values
    are stripped.

    Args:
        sep (str): key:value separator (Default : ':')

    Returns:
        dict: structured keys:values

    Examples:
        >>> s = '''
        ... a:val1
        ... b:
        ...     c:val3
        ...     d:
        ...         e ... : val5
        ...         f ... :val6
        ...     g:val7
        ... f: val8'''
        >>> s | parse_indented()
        {'a': 'val1', 'b': {'c': 'val3', 'd': {'e': 'val5', 'f': 'val6'}, 'g': 'val7'}, 'f': 'val8'}
        >>> s = '''
        ... a --> val1
        ... b --> val2'''
        >>> s | parse_indented(r'-->')
        {'a': 'val1', 'b': 'val2'}
    """
    @classmethod
    def op(cls, text, sep=r':', *args,**kwargs):
        indent_level = 0
        out = {}
        indent_node = {indent_level:out}
        dct = out
        prev_k = None
        # parse the text
        for line in cls._tolist(text):
            m = re.match(r'^(\s*)(\S.*)', line)
            if m:
                k,v = (re.split(sep,m.group(2)) + [''])[:2]
                indent = len(m.group(1))
                if indent < indent_level:
                    dct = indent_node.get(indent)
                    while dct is None:
                        indent -= 1
                        dct = indent_node.get(indent)
                    indent_level = indent
                    for ik in list(indent_node.keys()):
                        if ik > indent:
                            del indent_node[ik]
                elif indent > indent_level:
                    if prev_k is not None:
                        dct[prev_k] = {}
                        dct = dct[prev_k]
                    indent_node[indent] = dct
                    indent_level = indent
                k = index_normalize(k)
                v = v.strip()
                if k in dct:
                    prev_v = dct[k]
                    if isinstance(prev_v,dict):
                        dct[k]=[prev_v,{}]
                        dct = dct[k][-1]
                    elif isinstance(prev_v,str):
                        dct[k]=[prev_v,v]
                    else:
                        if isinstance(prev_v[0],str):
                            dct[k].append(v)
                        else:
                            dct[k].append({})
                            dct = dct[k][-1]
                    prev_k = None
                else:
                    dct[k]=v
                    prev_k = k
        return out

class parse_smart(TextOp):
    r"""Try to automatically parse a text

    It looks for key/value patterns, store found values in a dictionary.
    It tries to respect indents by creating sub-dictionaries.
    The keys are normalized (only keep ``A-Za-z0-9_``, the original key value
    is stored into the inner dict under the '_original_key' key),
    the values are stripped.

    Args:
        key_filter (func): a function that will receive a key before
        normalization and will return a new key string. The could be useful
        when a chapter title is too long. (Defaut : no filtering)

    Returns:
        dict: structured keys:values

    Examples:
        >>> s = '''
        ... Date/Time:       Wed Dec  2 09:51:17 NFT 2015
        ... Sequence Number: 156637
        ... Machine Id:      00F7B0114C00
        ...    Node Id:         xvio6
        ... Class:           H
        ... Type:            PERM
        ...    WPAR:            Global
        ...    Resource Name:   hdisk21
        ...       Resource Class:  disk
        ... Resource Type:   mpioapdisk
        ... Location:        U78AA.001.WZSHM0M-P1-C6-T1-W201400A0B8292A18-L13000000000000
        ...
        ... VPD:
        ...         Manufacturer................IBM
        ...         Machine Type and Model......1815      FAStT
        ...         ROS Level and ID............30393134
        ...         Serial Number...............
        ...         Device Specific.(Z0)........0000053245004032
        ...         Device Specific.(Z1)........
        ...
        ... Description
        ... DISK OPERATION ERROR
        ...
        ... Probable Causes
        ... DASD DEVICE
        ... '''
        >>> parsed = s >> parse_smart()
        >>> print(parsed.pretty())
        {   'class': 'H',
            'date_time': 'Wed Dec  2 09:51:17 NFT 2015',
            'description': ['DISK OPERATION ERROR'],
            'location': 'U78AA.001.WZSHM0M-P1-C6-T1-W201400A0B8292A18-L13000000000000',
            'machine_id': {   '_original_key': 'Machine Id',
                              'machine_id': '00F7B0114C00',
                              'node_id': 'xvio6'},
            'probable_causes': ['DASD DEVICE'],
            'resource_type': 'mpioapdisk',
            'sequence_number': '156637',
            'type': {   '_original_key': 'Type',
                        'resource_name': {   '_original_key': 'Resource Name',
                                             'resource_class': 'disk',
                                             'resource_name': 'hdisk21'},
                        'type': 'PERM',
                        'wpar': 'Global'},
            'vpd': {   '_original_key': 'VPD',
                       'device_specific_z0': '0000053245004032',
                       'device_specific_z1': '',
                       'machine_type_and_model': '1815      FAStT',
                       'manufacturer': 'IBM',
                       'ros_level_and_id': '30393134',
                       'serial_number': ''}}
        >>> print(parsed.vpd.device_specific_z0)
        0000053245004032
    """
    @classmethod
    def op(cls, text, key_filter=None, separators=None, *args,**kwargs):
        if separators is None:
            sep = r'[^a-zA-Z0-9_()\.-]{2,}|[:=]|\.{2,}|[_-]{3,}'
        else:
            sep = separators
        indent_level = 0
        out = {}
        indent_node = {indent_level:out}
        dct = out
        prev_k = None
        prev_original_k = None
        block_step = 1
        if key_filter is None:
            key_filter=lambda x:x
        # parse the text
        for line in cls._tolist(text):
            if not line.strip():
                block_step = 1
            else:
                m = re.match(r'^(\s*)(\S.*)', line)
                if m:
                    k,v = (re.split(sep,m.group(2),1) + [''])[:2]
                    original_k = k
                    v = v.strip()
                    indent = len(m.group(1))
                    #print indent,indent_level,block_step,line,'****',prev_k,k
                    if block_step==1:
                        if not v and re.search(r'\w+',k):
                            block_step = 2
                        else:
                            block_step = 0
                    elif block_step == 2:
                        if indent == indent_level:
                            block_k = prev_k
                            prev_k = index_normalize(key_filter(k))
                            prev_original_k = original_k
                            dct[block_k] = [m.group(2)]
                            block_step = 3
                            continue
                        else:
                            block_step = 0
                    elif block_step == 3:
                        if indent == indent_level:
                            dct[block_k].append(m.group(2))
                            prev_original_k = original_k
                            prev_k = index_normalize(key_filter(k))
                            continue
                        else:
                            block_step = 0

                    #print '-> ',indent,indent_level,block_step

                    if indent < indent_level:
                        dct = indent_node.get(indent)
                        while dct is None:
                            indent -= 1
                            dct = indent_node.get(indent)
                        indent_level = indent
                        for ik in list(indent_node.keys()):
                            if ik > indent:
                                del indent_node[ik]
                        if not v:
                            block_step = 2
                    elif indent > indent_level:
                        if prev_k is not None:
                            if prev_v:
                                dct[prev_k] = {prev_k:prev_v,
                                               '_original_key':prev_original_k}
                            else:
                                dct[prev_k] = {'_original_key':prev_original_k}
                            dct = dct[prev_k]
                        indent_node[indent] = dct
                        indent_level = indent

                    k = index_normalize(key_filter(k))
                    if k in dct:
                        prev_v = dct[k]
                        if isinstance(prev_v,dict):
                            dct[k]=[prev_v,{}]
                            dct = dct[k][-1]
                        elif isinstance(prev_v,str):
                            dct[k]=[prev_v,v]
                        else:
                            if isinstance(prev_v[0],str):
                                dct[k].append(v)
                            else:
                                dct[k].append({})
                                dct = dct[k][-1]
                        prev_k = None
                    else:
                        dct[k]=v
                        prev_k = k
                        prev_original_k = original_k
                        prev_v = v
        return out


class state_pattern(TextOp):
    r""" States and patterns parser

    This is a *state machine* parser :
    The main advantage is that it reads line-by-line the whole input text only once to collect all
    data you want into a multi-level dictionary. It uses patterns to select rules to be applied.
    It uses states to ensure only a set of rules are used against specific document sections.

    Args:
        states_patterns_desc (tupple) : descrption of states and patterns :
            see below for explaination
        reflags : re flags, ie re.I or re.M or re.I | re.M (Default : no flag)
        autostrip : before being stored, groupdict keys and values are stripped (Default : True)

    Returns:
        dict : parsed data from text

    |
    | **The states_patterns_desc :**

    It looks like this::

        ((<if state1>,<goto state1>,<pattern1>,<out data path1>,<out filter1>),
        ...
        (<if stateN>,<goto stateN>,<patternN>,<out data pathN>,<out filterN>))

    ``<if state>``
        is a string telling on what state(s) the pattern must be searched,
        one can specify several states with comma separated string or a tupple. if ``<if state>``
        is empty, the pattern will be searched for all lines.
        Note : at the beginning, the state is 'top'

    ``<goto state>``
        is a string corresponding to the new state if the pattern matches.
        use an empty string to not change the current state. One can use any string, usually,
        it corresponds to a specific section name of the document to parse where specific
        rules has to be used.
        if the pattern matches, no more rules are used for the current line except when you specify
        ``__continue__`` for the goto state. This is useful when you want to apply several rules on the same line.

    ``<pattern>``
        is a string or a re.regex to match a line of text.
        one should use named groups for selecting data, ex: ``(?P<key1>pattern)``

    ``<out data path>``
        is a string with a dot separator or a tuple telling where to place the groupdict
        from pattern maching process,
        The syntax is::

            '{contextkey1}.{contextkey2}. ... .{contextkeyN}'
            or
            ('{contextkey1}','{contextkey2}', ... ,'{contextkeyN}')
            or
            'key1.key2.keyN'
            or
            'key1.key2.keyN[]'
            or
            '{contextkey1}.{contextkey2}. ... .keyN[]'
            or
            '>context_dict_key'
            or
            '>>context_dict_key'
            or
            '>context_dict_key.{contextkey1}. ... .keyN'
            or
            '>>context_dict_key.{contextkey1}. ... .keyN'
            or
            None

        The contextdict (see after the definition) is used to format strings with ``{contextkeyN}`` syntax.
        instead of ``{contextkeyN}``, one can use a simple string to put data in a static path.

        Once the path fully formatted, let's say to ``key1.key2.keyN``, the parser will store the
        value into the result dictionnary at :
        ``{'key1':{'key2':{'keyN' : thevalue }}}``

        Example, Let's take the following data path ::

            data path : 'disks.{name}.{var}'

            if contextdict = {'name':'disk1','var':'size'}

            then the formatted data path is : 'disks.disk1.size',
            This means that the parsed data will be stored at :
            ``{'disks':{'disk1':{'size' : theparsedvalue depending on <out filter> }}}``


        One can use the string ``[]`` at the end of the path : the groupdict will be appended in a list
        ie : ``{'key1':{'key2':{'keyN' : [thevalue,...] }}}``

        if ``'>context_dict_key'`` is used, data is not store in parsed data but will be stored
        in context dict at ``context_dict_key`` key. by this way, you can differ the parsed date storage.
        To finally store to the parsed data use ``'<context_dict_key'`` for ``<out filter>``
        in some other rule.
        ``'>>context_dict_key'`` works like ``'>context_dict_key'`` but it updates data instead of
        replacing them (in others words : use ``>`` to start with an empty set of data, then
        use ``>>`` to update the data set).
        One can add dotted notation to complete data path: ``>>context_dict_key.{contextkey1}. ... .keyN``

        if None is used : nothing is stored

    ``<out filter>``
        is used to build the value to store,

        it could be :

            * None : no filter is applied, the re.MatchObject.groupdict() is stored
            * a dict : mainly to initalize the differed data set when using
              ``'>context_dict_key'`` in ``<out data path>``
            * ``'<context_dict_key'`` to store data from context dict at key ``context_dict_key``
            * a string : used as a format string with context dict, the formatted string is stored
            * a callable : to calculate the value to be stored and modify context dict if needed.
              the re.MatchObject and the context dict are given as arguments,
              it must return a tuple : the value to store AND the new context dict or None if unchanged

    **How the parser works :**

    You have a document where the syntax may change from one section to an another : You have just
    to give a name to these kind of sections : it will be your state names.
    The parser reads line by line the input text : For each line, it will look for the *first*
    matching rule from ``states_patterns_desc`` table, then will apply the rule.
    One rule has got 2 parts : the matching parameters, and the action parameters.

    Matching parameters:
        To match, a rule requires the parser to be at the specified state ``<if state>`` AND
        the line to be parsed must match the pattern ``<pattern>``. When the parser is at the first
        line, it has the default state ``top``. The pattern follow the
        standard python ``re`` module syntax. It is important to note that you must capture text
        you want to collect with the named group capture syntax, that is ``(?P<mydata>mypattern)``.
        By this way, the parser will store text corresponding to ``mypattern`` to a contextdict at
        the key ``mydata``.

    Action parameters:
        Once the rule matches, the action is to store ``<out filter>`` into the final dictionary at
        a specified ``<out data path>``.

    **Context dict :**

    The context dict is used within ``<out filter>`` and ``<out data path>``, it is a dictionary that
    is *PERSISTENT* during the whole parsing process :
    It is empty at the parsing beginning and will accumulate all captured pattern. For exemple, if
    a first rule pattern contains ``(?P<key1>.*),(?P<key2>.*)`` and matches the document line
    ``val1,val2``, the context dict will be ``{ 'key1' : 'val1', 'key2' : 'val2' }``. Then if a
    second rule pattern contains ``(?P<key2>.*):(?P<key3>.*)`` and matches the document line
    ``val4:val5`` then the context dict will be *UPDATED* to
    ``{ 'key1' : 'val1', 'key2' : 'val4', 'key3' : 'val5' }``.
    As you can see, the choice of the key names are *VERY IMPORTANT* in order to avoid collision
    across all the rules.

    Examples:

        >>> s = '''
        ... first name: Eric
        ... last name: Lapouyade'''
        >>> s | state_pattern( (('',None,'(?P<key>.*):(?P<val>.*)','{key}','{val}'),) )
        {'first_name': 'Eric', 'last_name': 'Lapouyade'}
        >>> s | state_pattern( (('',None,'(?P<key>.*):(?P<val>.*)','{key}',None),) ) #doctest: +NORMALIZE_WHITESPACE
        {'first_name': {'key': 'first name', 'val': 'Eric'},
        'last_name': {'key': 'last name', 'val': 'Lapouyade'}}
        >>> s | state_pattern((('',None,'(?P<key>.*):(?P<val>.*)','my.path.{key}','{val}'),))
        {'my': {'path': {'first_name': 'Eric', 'last_name': 'Lapouyade'}}}

        >>> s = '''Eric
        ... Guido'''
        >>> s | state_pattern( (('',None,'(?P<val>.*)','my.path.info[]','{val}'),) )
        {'my': {'path': {'info': ['Eric', 'Guido']}}}

        >>> s = '''
        ... Section 1
        ... ---------
        ...   email = ericdupo@gmail.com
        ...
        ... Section 2
        ... ---------
        ...   first name: Eric
        ...   last name: Dupont'''
        >>> s | state_pattern( (                                    #doctest: +NORMALIZE_WHITESPACE
        ... ('','section1','^Section 1',None,None),
        ... ('','section2','^Section 2',None,None),
        ... ('section1', '', '(?P<key>.*)=(?P<val>.*)', 'section1.{key}', '{val}'),
        ... ('section2', '', '(?P<key>.*):(?P<val>.*)', 'section2.{key}', '{val}')) )
        {'section1': {'email': 'ericdupo@gmail.com'},
        'section2': {'first_name': 'Eric', 'last_name': 'Dupont'}}

        >>> s = '''
        ... Disk states
        ... -----------
        ... name: c1t0d0s0
        ... state: good
        ... fs: /
        ... name: c1t0d0s4
        ... state: failed
        ... fs: /home
        ...
        ... '''
        >>> s | state_pattern( (                                    #doctest: +NORMALIZE_WHITESPACE
        ... ('top','disk',r'^Disk states',None,None),
        ... ('disk','top', r'^\s*$',None,None),
        ... ('disk', '', r'^name:(?P<diskname>.*)',None, None),
        ... ('disk', '', r'(?P<key>.*):(?P<val>.*)', 'disks.{diskname}.{key}', '{val}')) )
        {'disks': {'c1t0d0s0': {'state': 'good', 'fs': '/'},
        'c1t0d0s4': {'state': 'failed', 'fs': '/home'}}}

        >>> s = '''
        ... {
        ... name: c1t0d0s0
        ... state: good
        ... fs: /
        ... },
        ... {
        ... fs: /home
        ... name: c1t0d0s4
        ... }
        ... '''
        >>> s | state_pattern( (                                     #doctest: +NORMALIZE_WHITESPACE
        ... ('top','disk',r'{','>disk_info',{}),
        ... ('disk', '', r'(?P<key>.*):(?P<val>.*)', '>>disk_info.{key}', '{val}'),
        ... ('disk', 'top', r'}', 'disks.{disk_info[name]}', '<disk_info'),
        ... ) )
        {'disks': {'c1t0d0s0': {'name': 'c1t0d0s0', 'state': 'good', 'fs': '/'},
        'c1t0d0s4': {'fs': '/home', 'name': 'c1t0d0s4'}}}

        >>> s='firstname:Eric lastname=Lapouyade'
        >>> s | state_pattern((
        ... ('top','',r'firstname:(?P<val>\S+)','firstname','{val}'),
        ... ('top','',r'.*lastname=(?P<val>\S+)','lastname','{val}'),
        ... ))
        {'firstname': 'Eric'}

        >>> s='firstname:Eric lastname=Lapouyade'
        >>> s | state_pattern((
        ... ('top','__continue__',r'firstname:(?P<val>\S+)','firstname','{val}'),
        ... ('top','',r'.*lastname=(?P<val>\S+)','lastname','{val}'),
        ... ))
        {'firstname': 'Eric', 'lastname': 'Lapouyade'}

    """

    @classmethod
    def op(cls,text, states_patterns_desc, reflags=0, autostrip=True,**kwargs):
        state = 'top'
        root_data = {}
        groups_context = {}
        states_patterns = []

        #check states_patterns_desc is a correct tuple/list of tuples/lists
        if not isinstance(states_patterns_desc,(list,tuple)):
            raise ParsingError('states_patterns_desc must contains a tuple/list of tuples/lists')
        if not states_patterns_desc or not states_patterns_desc[0]:
            raise ParsingError('states_patterns_desc must not be empty')
        if not isinstance(states_patterns_desc[0],(list,tuple)):
            raise ParsingError('states_patterns_desc must contains a tuple/list of tuples/lists : one level of parenthesis or a coma is missing somewhere.')
        if len(states_patterns_desc[0]) != 5:
            raise ParsingError('states_patterns_desc subtuple must contain 5 elements : ifstate, gotostate, pattern, datapath and outfilter')

        #normalizing states_patterns_desc :
        for ifstate, gotostate, pattern, datapath, outfilter in states_patterns_desc:
            if not ifstate:
                ifstate = ()
            else:
                if isinstance(ifstate, str):
                    ifstate = ifstate.split(',')
            if isinstance(pattern, str):
                pattern = re.compile(pattern,reflags)
            if isinstance(datapath, str):
                if not datapath:
                    datapath = []
                else:
                    datapath = datapath.split('.')
            states_patterns.append((ifstate, gotostate, pattern, datapath, outfilter))

        # parse the text
        for line in cls._tolist(text):
            logger.debug('state:%10s, line = %s',state, line)
            for ifstate, gotostate, pattern, datapath, outfilter in states_patterns:
                logger.debug('  ? %10s "%10s" r\'%s\' "%s" "%s"',ifstate, gotostate, pattern.pattern, datapath, outfilter)
                if not ifstate or state in ifstate:
                    m=pattern.match(line)
                    if m:
                        g = m.groupdict()
                        if autostrip:
                            g = dict([ (k,v.strip() if isinstance(v,str) else v) for k,v in list(m.groupdict().items()) ])
                        # replace None values by NoAttr so dotted notation can be used when a named group does not match anything.
                        for k,v in g.items():
                            if v is None:
                                g[k] = NoAttr
                        groups_context.update(g,_ifstate=ifstate,_gotostate=gotostate,_state=state)
                        logger.debug('  -> OK')
                        logger.debug('    context = %s',groups_context)
                        if datapath is not None:
                            if datapath[0].startswith('>'):
                                k = datapath[0][1:].strip()
                                if k.startswith('>'):
                                    k = k[1:]
                                    data = groups_context.setdefault(k,textops.DefaultDict(lambda k:'_%s_not_found' % k,{}))
                                else:
                                    groups_context[k] = textops.DefaultDict(lambda k:'_%s_not_found' % k,{})
                                    data = groups_context[k]
                                datapath = datapath[1:]
                            else:
                                data = root_data

                            for p in datapath:
                                p = dformat(p,groups_context,context_key_not_found)
                                prev_data = data
                                if p[-2:] == '[]':
                                    p = p[:-2]
                                    p = index_normalize(p)
                                    if p not in data:
                                        data[p] = []
                                    data = data[p]
                                else:
                                    p = index_normalize(p)
                                    if p not in data:
                                        data[p] = {}
                                    data = data[p]

                            if isinstance(outfilter, collections.abc.Callable):
                                g,new_groups_context=outfilter(m,groups_context)
                                if new_groups_context is not None:
                                    groups_context = new_groups_context
                            elif isinstance(outfilter, str):
                                if outfilter.startswith('<'):
                                    g = groups_context.get(outfilter[1:].strip(),{})
                                else:
                                    g=dformat(outfilter,groups_context,context_key_not_found)
                            elif isinstance(outfilter, dict):
                                g = outfilter

                            if isinstance(data,list):
                                data.append(g)
                            else:
                                if isinstance(g,dict):
                                    data.update(g)
                                else:
                                    prev_data[p] = g
                        if gotostate == '__continue__':
                            continue
                        if gotostate:
                            state = gotostate
                        break
        return root_data



# -*- coding: utf-8 -*-
#
# Created : 2015-04-03
#
# @author: Eric Lapouyade
#
""" This module gathers text operations to be run on a string """

from textops import TextOp
import re
import types
from functools import reduce

class PyStrWrapper(object):
    """ This is a help to be able to use python string method with piped notation """

    def __getattr__(self,attr):
        if not attr.startswith('_'):
            op = TextOp()
            op.ops=[[attr, (), {}]]
            op.op = attr
        else:
            raise AttributeError()
        return op

strop = PyStrWrapper()

class length(TextOp):
    r""" Returns the length of a string, list or generator

    Returns:
        int: length of the string

    Examples:

        >>> s='this is a string'
        >>> s | length()
        16
        >>> s=StrExt(s)
        >>> s.length()
        16
        >>> ['a','b','c'] | length()
        3
        >>> def mygenerator():yield 3; yield 2
        >>> mygenerator() | length()
        2
    """
    @classmethod
    def op(cls,text,*args,**kwargs):
        if isinstance(text, types.GeneratorType):
            return reduce(lambda x,y:x+1,text,0)
        return len(text)

class echo(TextOp):
    r""" identity operation

    it returns the same text, except that is uses textops Extended classes (StrExt, ListExt ...).
    This could be usefull in some cases to access str methods (upper, replace, ...) just after a pipe.

    Returns:
        int: length of the string

    Examples:

        >>> s='this is a string'
        >>> type(s)
        <class 'str'>
        >>> t=s | echo()
        >>> type(t)
        <class 'textops.base.StrExt'>
        >>> s.upper()
        'THIS IS A STRING'
        >>> s | upper()
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        NameError: name 'upper' is not defined
        >>> s | echo().upper()
        'THIS IS A STRING'
        >>> s | strop.upper()
        'THIS IS A STRING'
    """
    fn = staticmethod(lambda text: text)

class splitln(TextOp):
    r""" Transforms a string with newlines into a list of lines

    It uses python str.splitlines() : newline separator can be \\n or \\r or both. They are removed
    during the process.

    Returns:
        list: The splitted text

    Example:

        >>> s='this is\na multi-line\nstring'
        >>> s | splitln()
        ['this is', 'a multi-line', 'string']
    """
    @classmethod
    def op(cls,text,*args,**kwargs):
        return cls._tolist(text)

class matches(TextOp):
    r""" Tests whether a pattern is present or not

    Uses re.match() to match a pattern against the string.

    Args:
        pattern (str): a regular expression string

    Returns:
        re.RegexObject: The pattern found

    Note:
        Be careful : the pattern is tested from the beginning of the string, the pattern is NOT searched somewhere in the middle of the string.

    Examples:

        >>> state=StrExt('good')
        >>> print('OK' if state.matches(r'good|not_present|charging') else 'CRITICAL')
        OK
        >>> state=StrExt('looks like all is good')
        >>> print('OK' if state.matches(r'good|not_present|charging') else 'CRITICAL')
        CRITICAL
        >>> print('OK' if state.matches(r'.*(good|not_present|charging)') else 'CRITICAL')
        OK
        >>> state=StrExt('Error')
        >>> print('OK' if state.matches(r'good|not_present|charging') else 'CRITICAL')
        CRITICAL

    """
    @classmethod
    def op(cls,text,pattern,*args,**kwargs):
        return re.match(pattern,text,*args,**kwargs)

class searches(TextOp):
    r""" Search a pattern

    Uses re.search() to find a pattern in the string.

    Args:
        pattern (str): a regular expression string

    Returns:
        re.RegexObject: The pattern found

    Examples:

        >>> state=StrExt('good')
        >>> print('OK' if state.searches(r'good|not_present|charging') else 'CRITICAL')
        OK
        >>> state=StrExt('looks like all is good')
        >>> print('OK' if state.searches(r'good|not_present|charging') else 'CRITICAL')
        OK
        >>> print('OK' if state.searches(r'.*(good|not_present|charging)') else 'CRITICAL')
        OK
        >>> state=StrExt('Error')
        >>> print('OK' if state.searches(r'good|not_present|charging') else 'CRITICAL')
        CRITICAL

    """
    @classmethod
    def op(cls,text,pattern,*args,**kwargs):
        return re.search(pattern,text,*args,**kwargs)



class StrOp(TextOp):
    @classmethod
    def op(cls,text,*args,**kwargs):
        if ( ( isinstance(text, str) and '\n' in text ) or
             (isinstance(text, bytes) and b'\n' in text) ):
            text = cls._tolist(text)
        if isinstance(text, (str,bytes)):
            return cls.fn(text,*args,**kwargs)
        elif isinstance(text, list):
            return [ cls.fn(line,*args,**kwargs) for line in text ]
        return cls.gop(text,*args,**kwargs)
    @classmethod
    def gop(cls,text,*args,**kwargs):
        for line in text:
            yield cls.fn(line,*args,**kwargs)

class cut(StrOp):
    r""" Extract columns from a string or a list of strings

    This works like the unix shell command 'cut'. It uses :meth:`str.split` function.

        * if the input is a simple string, cut() will return a list of strings
          representing the splitted input string.
        * if the input is a list of strings or a string with newlines, cut() will return
          a list of list of strings : each line of the input will splitted and put in a list.
        * if only one column is extracted, one level of list is removed.

    Args:
        sep (str): a string as a column separator, default is None : this means 'any kind of spaces'
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        A string, a list of strings or a list of list of strings

    Examples:

        >>> s='col1 col2 col3'
        >>> s | cut()
        ['col1', 'col2', 'col3']
        >>> s | cut(col=1)
        'col2'
        >>> s | cut(col='1,2,10',default='N/A')
        ['col2', 'col3', 'N/A']
        >>> s='col1.1 col1.2 col1.3\ncol2.1 col2.2 col2.3'
        >>> s | cut()
        [['col1.1', 'col1.2', 'col1.3'], ['col2.1', 'col2.2', 'col2.3']]
        >>> s | cut(col=1)
        ['col1.2', 'col2.2']
        >>> s | cut(col='0,1')
        [['col1.1', 'col1.2'], ['col2.1', 'col2.2']]
        >>> s | cut(col=[1,2])
        [['col1.2', 'col1.3'], ['col2.2', 'col2.3']]
        >>> s='col1.1 | col1.2 |  col1.3\ncol2.1 | col2.2 | col2.3'
        >>> s | cut()
        [['col1.1', '|', 'col1.2', '|', 'col1.3'], ['col2.1', '|', 'col2.2', '|', 'col2.3']]
        >>> s | cut(sep=' | ')
        [['col1.1', 'col1.2', ' col1.3'], ['col2.1', 'col2.2', 'col2.3']]
    """
    sep_is_regex = False
    flags = 0

    @classmethod
    def split(cls, text, sep, maxsplit=-1,*args,**kwargs):
        return text.split(sep,maxsplit)

    @classmethod
    def fn(cls, text, sep=None, col=None, default='', *args,**kwargs):
        if cls.sep_is_regex:
            if isinstance(sep, str):
                sep = re.compile(sep,cls.flags)
        if isinstance(col, str):
            col = [int(i) for i in col.split(',')]
        if isinstance(col,(list,tuple)):
            nbcol = len(col)
            line_cols = cls.split(text,sep, *args,**kwargs)
            nblinecol = len(line_cols)
            return [ line_cols[c] if c < nblinecol else default for c in col ]
        else:
            line_cols = cls.split(text,sep, *args,**kwargs)
            if line_cols is None:
                return default
            nblinecol = len(line_cols)
            if col == None:
                return line_cols
            elif col < nblinecol:
                return line_cols[col]
            else:
                return default

class cutre(cut):
    r""" Extract columns from a string or a list of strings with re.split()

    This works like the unix shell command 'cut'. It uses :func:`re.split` function.

        * if the input is a simple string, cutre() will return a list of strings
          representing the splitted input string.
        * if the input is a list of strings or a string with newlines, cut() will return
          a list of list of strings : each line of the input will splitted and put in a list.
        * if only one column is extracted, one level of list is removed.

    Args:
        sep (str or re.RegexObject): a regular expression string or object as a column separator
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        A string, a list of strings or a list of list of strings

    Examples:

        >>> s='col1.1 | col1.2 | col1.3\ncol2.1 | col2.2 | col2.3'
        >>> print(s)
        col1.1 | col1.2 | col1.3
        col2.1 | col2.2 | col2.3
        >>> s | cutre(r'\s+')
        [['col1.1', '|', 'col1.2', '|', 'col1.3'], ['col2.1', '|', 'col2.2', '|', 'col2.3']]
        >>> s | cutre(r'[\s|]+')
        [['col1.1', 'col1.2', 'col1.3'], ['col2.1', 'col2.2', 'col2.3']]
        >>> s | cutre(r'[\s|]+','0,2,4','-')
        [['col1.1', 'col1.3', '-'], ['col2.1', 'col2.3', '-']]
        >>> mysep = re.compile(r'[\s|]+')
        >>> s | cutre(mysep)
        [['col1.1', 'col1.2', 'col1.3'], ['col2.1', 'col2.2', 'col2.3']]
    """
    sep_is_regex = True

    @classmethod
    def split(cls, text, sep, maxsplit=0, *args,**kwargs):
        return sep.split(text,maxsplit)

class cutca(cut):
    r""" Extract columns from a string or a list of strings through pattern capture

    This works like :class:`textops.cutre` except it needs a pattern having parenthesis to capture column.
    It uses :func:`re.match` for capture, this means the pattern must start at line beginning.

        * if the input is a simple string, cutca() will return a list of strings
          representing the splitted input string.
        * if the input is a list of strings or a string with newlines, cut() will return
          a list of list of strings : each line of the input will splitted and put in a list.
        * if only one column is extracted, one level of list is removed.

    Args:
        sep (str or re.RegexObject): a regular expression string or object having capture parenthesis
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        a list of strings or a list of list of strings

    Examples:

        >>> s='-col1- =col2= _col3_'
        >>> s | cutca(r'[^-]*-([^-]*)-[^=]*=([^=]*)=[^_]*_([^_]*)_')
        ['col1', 'col2', 'col3']
        >>> s=['-col1- =col2= _col3_','-col11- =col22= _col33_']
        >>> s | cutca(r'[^-]*-([^-]*)-[^=]*=([^=]*)=[^_]*_([^_]*)_','0,2,4','not present')
        [['col1', 'col3', 'not present'], ['col11', 'col33', 'not present']]
    """
    sep_is_regex = True

    @classmethod
    def split(cls, text, sep, *args,**kwargs):
        m = sep.match(text)
        return list(m.groups()) if m else []

class cutm(cut):
    r""" Extract exactly one column by using :func:`re.match`

    It returns the matched pattern. Beware : the pattern must match the beginning of the line.
    One may use capture parenthesis to only return a part of the found pattern.

        * if the input is a simple string, :class:`textops.cutm` will return a strings
          representing the captured substring.
        * if the input is a list of strings or a string with newlines, :class:`textops.cutm` will
          return a list of captured substring.

    Args:
        sep (str or re.RegexObject): a regular expression string or object having capture parenthesis
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        a list of strings or a list of list of strings

    Examples:

        >>> s='-col1- =col2= _col3_'
        >>> s | cutm(r'[^=]*=[^=]*=')
        '-col1- =col2='
        >>> s | cutm(r'=[^=]*=')
        ''
        >>> s | cutm(r'[^=]*=([^=]*)=')
        'col2'
        >>> s=['-col1- =col2= _col3_','-col11- =col22= _col33_']
        >>> s | cutm(r'[^-]*-([^-]*)-')
        ['col1', 'col11']
        >>> s | cutm(r'[^-]*-(badpattern)-',default='-')
        ['-', '-']
    """
    sep_is_regex = True

    @classmethod
    def split(cls, text, sep, *args,**kwargs):
        m = sep.match(text)
        if not m:
            return None
        g = m.groups()
        if not g:
            return m.group(0)
        return g[0]

class cutmi(cutm):
    r""" Extract exactly one column by using :func:`re.match` (case insensitive)

    This works like :class:`textops.cutm` except it is case insensitive.

        * if the input is a simple string, :class:`textops.cutmi` will return a strings
          representing the captured substring.
        * if the input is a list of strings or a string with newlines, :class:`textops.cutmi` will
          return a list of captured substring.

    Args:
        sep (str or re.RegexObject): a regular expression string or object having capture parenthesis
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        a list of strings or a list of list of strings

    Examples:

        >>> s='-col1- =col2= _col3_'
        >>> s | cutm(r'.*(COL\d+)',default='no found')
        'no found'
        >>> s='-col1- =col2= _col3_'
        >>> s | cutmi(r'.*(COL\d+)',default='no found')  #as .* is the longest possible, only last column is extracted
        'col3'
    """
    flags = re.IGNORECASE

class cuts(cut):
    r""" Extract exactly one column by using :func:`re.search`

    This works like :class:`textops.cutm` except it searches the first occurence of the pattern in
    the string. One may use capture parenthesis to only return a part of the found pattern.

        * if the input is a simple string, :class:`textops.cuts` will return a strings
          representing the captured substring.
        * if the input is a list of strings or a string with newlines, :class:`textops.cuts` will
          return a list of captured substring.

    Args:
        sep (str or re.RegexObject): a regular expression string or object having capture parenthesis
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        a list of strings or a list of list of strings

    Examples:

        >>> s='-col1- =col2= _col3_'
        >>> s | cuts(r'_[^_]*_')
        '_col3_'
        >>> s | cuts(r'_([^_]*)_')
        'col3'
        >>> s=['-col1- =col2= _col3_','-col11- =col22= _col33_']
        >>> s | cuts(r'_([^_]*)_')
        ['col3', 'col33']
    """
    sep_is_regex = True

    @classmethod
    def split(cls, text, sep, *args,**kwargs):
        m = sep.search(text)
        if not m:
            return None
        g = m.groups()
        if not g:
            return m.group(0)
        return g[0]

class cutsi(cuts):
    r""" Extract exactly one column by using :func:`re.search` (case insensitive)

    This works like :class:`textops.cuts` except it is case insensitive.

    Args:
        sep (str or re.RegexObject): a regular expression string or object having capture parenthesis
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        a list of strings or a list of list of strings

    Examples:

        >>> s='-col1- =col2= _col3_'
        >>> s | cuts(r'_(COL[^_]*)_')
        ''
        >>> s='-col1- =col2= _col3_'
        >>> s | cutsi(r'_(COL[^_]*)_')
        'col3'
    """
    flags = re.IGNORECASE

class cutsa(cut):
    r""" Extract all columns having the specified pattern.

    It uses :func:`re.finditer` to find all occurences of the pattern.

        * if the input is a simple string, :class:`textops.cutfa` will return a list of found strings.
        * if the input is a list of strings or a string with newlines, :class:`textops.cutfa` will
          return a list of list of found string.

    Args:
        sep (str or re.RegexObject): a regular expression string or object having capture parenthesis
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        a list of strings or a list of list of strings

    Examples:

        >>> s='-col1- =col2= _col3_'
        >>> s | cutsa(r'col\d+')
        ['col1', 'col2', 'col3']
        >>> s | cutsa(r'col(\d+)')
        ['1', '2', '3']
        >>> s=['-col1- =col2= _col3_','-col11- =col22= _col33_']
        >>> s | cutsa(r'col\d+')
        [['col1', 'col2', 'col3'], ['col11', 'col22', 'col33']]
    """
    sep_is_regex = True

    @classmethod
    def split(cls, text, sep, *args,**kwargs):
        cols=[]
        for m in sep.finditer(text):
            g = m.groups()
            cols.append(g[0] if g else m.group(0))
        return cols

class cutsai(cutsa):
    r""" Extract all columns having the specified pattern. (case insensitive)

    It works like :class:`textops.cutsa` but is case insensitive if the pattern is given as a string.

    Args:
        sep (str or re.RegexObject): a regular expression string or object having capture parenthesis
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        a list of strings or a list of list of strings

    Examples:

        >>> s='-col1- =col2= _col3_'
        >>> s | cutsa(r'COL\d+')
        []
        >>> s | cutsai(r'COL\d+')
        ['col1', 'col2', 'col3']
        >>> s | cutsai(r'COL(\d+)')
        ['1', '2', '3']
        >>> s=['-col1- =col2= _col3_','-col11- =col22= _col33_']
        >>> s | cutsai(r'COL\d+')
        [['col1', 'col2', 'col3'], ['col11', 'col22', 'col33']]
    """
    flags = re.IGNORECASE

class cutdct(cut):
    r""" Extract columns from a string or a list of strings through pattern capture

    This works like :class:`textops.cutca` except it needs a pattern having *named* parenthesis to capture column.

        * if the input is a simple string, cutca() will return a list of strings
          representing the splitted input string.
        * if the input is a list of strings or a string with newlines, cut() will return
          a list of list of strings : each line of the input will splitted and put in a list.
        * if only one column is extracted, one level of list is removed.

    Args:
        sep (str or re.RegexObject): a regular expression string or object
            having *named* capture parenthesis
        col (int or list of int or str) : specify one or many columns you want to get back,
            You can specify :

            * an int as a single column number (starting with 0)
            * a list of int as the list of colmun
            * a string containing a comma separated list of int
            * None (default value) for all columns

        default (str): A string to display when requesting a column that does not exist

    Returns:
        A string, a list of strings or a list of list of strings

    Examples:

        >>> s='item="col1" count="col2" price="col3"'
        >>> s | cutdct(r'item="(?P<item>[^"]*)" count="(?P<i_count>[^"]*)" price="(?P<i_price>[^"]*)"')
        {'item': 'col1', 'i_count': 'col2', 'i_price': 'col3'}
        >>> s='item="col1" count="col2" price="col3"\nitem="col11" count="col22" price="col33"'
        >>> s | cutdct(r'item="(?P<item>[^"]*)" count="(?P<i_count>[^"]*)" price="(?P<i_price>[^"]*)"') # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        [{'item': 'col1', 'i_count': 'col2', 'i_price': 'col3'},
        {'item': 'col11', 'i_count': 'col22', 'i_price': 'col33'}]
    """
    sep_is_regex = True

    @classmethod
    def split(cls, text, sep, *args,**kwargs):
        m = sep.match(text)
        return m.groupdict() if m else {}

class cutkv(cut):
    r""" Extract columns from a string or a list of strings through pattern capture

    This works like :class:`textops.cutdct` except it return a dict where the key is the one captured with the
    name given in parameter 'key_name', and where the value is the full dict of captured values.
    The interest is to merge informations into a bigger dict : see ``merge_dicts()``

    Args:
        sep (str or re.RegexObject): a regular expression string or object
            having *named* capture parenthesis
        key_name (str) : specify the named capture to use as the key for the returned dict
            Default value is 'key'

    Note:
        ``key_name=`` must be specified (not a positionnal parameter)

    Returns:
        A dict or a list of dict

    Examples:

        >>> s='item="col1" count="col2" price="col3"'
        >>> pattern=r'item="(?P<item>[^"]*)" count="(?P<i_count>[^"]*)" price="(?P<i_price>[^"]*)"'
        >>> s | cutkv(pattern,key_name='item')
        {'col1': {'item': 'col1', 'i_count': 'col2', 'i_price': 'col3'}}
        >>> s='item="col1" count="col2" price="col3"\nitem="col11" count="col22" price="col33"'
        >>> s | cutkv(pattern,key_name='item')                                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        [{'col1': {'item': 'col1', 'i_count': 'col2', 'i_price': 'col3'}},
        {'col11': {'item': 'col11', 'i_count': 'col22', 'i_price': 'col33'}}]

    """
    sep_is_regex = True

    @classmethod
    def split(cls, text, sep, key_name = 'key', *args,**kwargs):
        # Use named 'key_name' parameter, not postionnal
        m = sep.match(text)
        if m:
            dct = m.groupdict()
            kv = dct.get(key_name)
            if kv:
                return { kv : dct }
        return {}


# -*- coding: utf-8 -*-
#
# Created : 2017-03-23
#
# @author: Eric Lapouyade
#
import inspect
from textops import TextOp
from datetime import datetime
import os

modules = [ f[:-3] for f in os.listdir('../textops/ops') if not f.endswith('.pyc') and not f.startswith('_') ]

for mod_name in modules:
    print(mod_name,'...')

    m = __import__('textops.ops.%s' % mod_name, fromlist=[''])

    out = """..
   Generated: %s

   @author : Eric Lapouyade

""" % datetime.now().strftime('%c')

    out += '=' * len(mod_name)
    out += '\n' + mod_name
    out += '\n' + '=' * len(mod_name)
    out += '\n'
    out += '\n' + '.. automodule:: textops.ops.%s' % mod_name
    out += '\n' + '.. currentmodule:: textops'
    out += '\n'

    for cls_name,cls in sorted(vars(m).items(),key=lambda x:x[0]):
        if isinstance(cls,type):
            if issubclass(cls,TextOp) and cls is not TextOp:
                meth = None
                addvarargs = False
                if hasattr(cls,'fn'):
                    meth = cls.fn
                elif hasattr(cls, 'testline'):
                    meth = cls.testline
                    addvarargs=True
                else:
                    meth = cls.op

                try:
                    if callable(meth):
                        margs = inspect.getargspec(meth)
                        args = margs.args
                        defaults = margs.defaults
                        if defaults:
                            for i,d in enumerate(reversed(margs.defaults),start=1):
                                if d == 9223372036854775807:
                                    args[-i] = '%s=sys.maxint' % args[-i]
                                else:
                                    args[-i] = '%s=%r' % (args[-i],d)

                        out += '\n' + cls_name
                        out += '\n' + '-' * len(cls_name)
                        out += '\n' + '   .. autoclass:: %s(%s%s)' % (cls_name,', '.join(args[2:]),', *args, **kwargs' if addvarargs else '')
                        out += '\n'
                except TypeError:
                    pass
    with open('%s.rst' % mod_name,'w') as fh:
        fh.write(out)


from setuptools import setup,find_packages
import os
import re

def read(*names):
    values = []
    for filename in names:
        if os.path.isfile(filename):
            try:
                with open(filename) as fd:
                    values.append(fd.read())
            except:
                values.append('')
        else:
            values.append('')
    return values


long_description = """{0}

News
====

{1}
""".format(*read('README.rst', 'CHANGES.rst'))

def get_version(pkg):
    path = os.path.join(os.path.dirname(__file__),pkg,'__init__.py')
    with open(path) as fh:
        m = re.search(r'^__version__\s*=\s*[\'"]([^\'"]+)[\'"]',fh.read(),re.M)
    if m:
        return m.group(1)
    raise RuntimeError("Unable to find __version__ string in %s." % path)

setup(name='python-textops3',
      version=get_version('textops'),
      description='Python text operations module',
      long_description=long_description,
      classifiers=[
          "Intended Audience :: Developers",
          "Development Status :: 4 - Beta",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
      ],
      keywords='textops',
      url='https://github.com/elapouya/python-textops3',
      author='Eric Lapouyade',
      author_email='elapouya@gmail.com',
      license='LGPL 2.1',
      packages=find_packages(),
      install_requires=['addicted3',
                        'python-dateutil',
                        'python-slugify',
                        'chardet'],
      extras_require={'docs': ['Sphinx', 'sphinxcontrib-napoleon']},
      eager_resources=['docs'],
      zip_safe=False)



# -*- coding: utf-8 -*-
#
# Created : 2015-04-03
#
# @author: Eric Lapouyade
#
""" This modules provides casting features, that is to force the output type """

from textops import TextOp, pp, stru
import dateutil.parser
from slugify import slugify

class tostr(TextOp):
    r""" Convert the result to a string

    | If the input text is a list or a generator, it will join all the lines with a newline.
    | If the input text is None, None is NOT converted to a string : None is returned

    Args:
        join_str (str): the join string to apply on list or generator (Default : newline)
        return_if_none (str): the object to return if the input text is None (Default : None)

    Returns:
        str or None: converted result as a string

    Examples:

        >>> 'line1\nline2' | tostr()
        'line1\nline2'
        >>> ['line1','line2'] | tostr()
        'line1\nline2'
        >>> ['line1','line2'] | tostr('---')
        'line1---line2'
        >>> def g(): yield 'hello';yield 'world'
        ...
        >>> g()|tostr()
        'hello\nworld'
        >>> type(None | tostr())
        <class 'NoneType'>
        >>> None | tostr(return_if_none='N/A')
        'N/A'
    """
    @classmethod
    def fn(cls, text, join_str='\n', return_if_none=None,*args,**kwargs):
        return TextOp.make_string(text,join_str,return_if_none)

class tostre(TextOp):
    r""" Convert the result to a string

    | If the input text is a list or a generator, it will join all the lines with a newline.
    | If the input text is None, None is converted to an empty string.

    Args:
        join_str (str): the join string to apply on list or generator (Default : newline)
        return_if_none (str): the object to return if the input text is None (Default : empty string)

    Returns:
        str: converted result as a string

    Examples:

        >>> ['line1','line2'] | tostre()
        'line1\nline2'
        >>> type(None | tostre())
        <class 'textops.base.StrExt'>
        >>> None | tostre()
        ''
    """
    @classmethod
    def fn(cls, text, join_str='\n', return_if_none='',*args,**kwargs):
        return TextOp.make_string(text,join_str,return_if_none)

class tolist(TextOp):
    r""" Convert the result to a list

    | If the input text is a string, it is put in a list.
    | If the input text is a list : nothing is done.
    | If the input text is a generator : it is converted into a list

    Args:
        return_if_none (str): the object to return if the input text is None (Default : None)

    Returns:
        str: converted result as a string

    Examples:
        >>> 'hello' | tolist()
        ['hello']
        >>> ['hello','world'] | tolist()
        ['hello', 'world']
        >>> type(None|tolist())
        <class 'NoneType'>
        >>> def g(): yield 'hello'
        ...
        >>> g()|tolist()
        ['hello']
    """
    @classmethod
    def fn(cls, text, return_if_none=None,*args,**kwargs):
        return TextOp.make_list(text,return_if_none)

class toliste(TextOp):
    r""" Convert the result to a list

    | If the input text is a string, it is put in a list.
    | If the input text is a list : nothing is done.
    | If the input text is a generator : it is converted into a list

    Args:
        return_if_none (str): the object to return if the input text is None (Default : empty list)

    Returns:
        str: converted result as a string

    Examples:
        >>> 'hello' | toliste()
        ['hello']
        >>> type(None|toliste())
        <class 'textops.base.ListExt'>
        >>> None|toliste()
        []
    """
    @classmethod
    def fn(cls, text, return_if_none=None,*args,**kwargs):
        if return_if_none is None:
            return_if_none = []
        return TextOp.make_list(text,return_if_none)

class toint(TextOp):
    r""" Convert the result to an integer

    Returns:
        str or list: converted result as an int or list of int

    Examples:
        >>> '53' | toint()
        53
        >>> 'not an int' | toint()
        0
        >>> '3.14' | toint()
        3
        >>> '3e3' | toint()
        3000
        >>> ['53','not an int','3.14'] | toint()
        [53, 0, 3]
    """
    @classmethod
    def fn(cls, text,*args,**kwargs):
        return TextOp.make_int(text)

class tofloat(TextOp):
    r""" Convert the result to a float

    Returns:
        str or list: converted result as an int or list of int

    Examples:
        >>> '53' | tofloat()
        53.0
        >>> 'not a float' | tofloat()
        0.0
        >>> '3.14' | tofloat()
        3.14
        >>> '3e3' | tofloat()
        3000.0
        >>> ['53','not an int','3.14'] | tofloat()
        [53.0, 0.0, 3.14]
    """
    @classmethod
    def fn(cls, text,*args,**kwargs):
        return TextOp.make_float(text)

class todatetime(TextOp):
    r""" Convert the result to a datetime python object

    Returns:
        datetime: converted result as a datetime python object

    Examples:
        >>> '2015-10-28' | todatetime()
        datetime.datetime(2015, 10, 28, 0, 0)
        >>> '2015-10-28 22:33:00' | todatetime()
        datetime.datetime(2015, 10, 28, 22, 33)
        >>> '2015-10-28 22:33:44' | todatetime()
        datetime.datetime(2015, 10, 28, 22, 33, 44)
        >>> '2014-07-08T09:02:21.377' | todatetime()
        datetime.datetime(2014, 7, 8, 9, 2, 21, 377000)
        >>> '28-10-2015' | todatetime()
        datetime.datetime(2015, 10, 28, 0, 0)
        >>> '10-28-2015' | todatetime()
        datetime.datetime(2015, 10, 28, 0, 0)
        >>> '10-11-2015' | todatetime()
        datetime.datetime(2015, 10, 11, 0, 0)
    """
    @classmethod
    def fn(cls, text,*args,**kwargs):
        return dateutil.parser.parse(text)

class toslug(TextOp):
    r""" Convert a string to a slug

    Returns:
        str: a slug

    Examples:
        >>> 'this is my article' | toslug()
        'this-is-my-article'
        >>> 'this%%% is### my___article' | toslug()
        'this-is-my-article'
    """
    @classmethod
    def fn(cls, text,*args,**kwargs):
        return slugify(text)

class todict(TextOp):
    r"""Converts list or 2 items-tuples into dict
    
    Returns:
        dict: Converted result as a dict

    Examples:
        
    >>> [ ('a',1), ('b',2), ('c',3) ] | echo().todict()
    {'a': 1, 'b': 2, 'c': 3}
    """         
    @classmethod
    def fn(cls, text,*args,**kwargs):
        return TextOp.make_dict(text)


class tonull(TextOp):
    r"""Consume generator if any and then return nothing (aka None)

    Examples:

    >>> [ 1,2,3 ] | echo().tonull()

    """
    @classmethod
    def fn(cls, text, *args, **kwargs):
        TextOp.consume(text)

class pretty(TextOp):
    r"""Pretty format the input text
    
    Returns:
        str: Converted result as a pretty string ( uses :meth:`pprint.PrettyPrinter.pformat` )

    Examples:
        
    >>> s = '''
    ... a:val1
    ... b:
    ...     c:val3
    ...     d:
    ...         e ... : val5
    ...         f ... :val6
    ...     g:val7
    ... f: val8'''
    >>> print(s | parse_indented())
    {'a': 'val1', 'b': {'c': 'val3', 'd': {'e': 'val5', 'f': 'val6'}, 'g': 'val7'}, 'f': 'val8'}
    >>> print(s | parse_indented().pretty())
    {   'a': 'val1',
        'b': {'c': 'val3', 'd': {'e': 'val5', 'f': 'val6'}, 'g': 'val7'},
        'f': 'val8'}
    """         
    @classmethod
    def fn(cls, text,*args,**kwargs):
        return pp.pformat(text)



from .cast import *
from .fileops import *
from .strops import *
from .listops import *
from .runops import *
from .parse import *
from .wrapops import *
from .recode import *

SPLIT_SEP_NONE = 0
SPLIT_SEP_BEGIN = 1
SPLIT_SEP_END = 2
